"""Implementation of arrow storage for time series."""

import atexit
import shutil
from datetime import datetime
from pathlib import Path
from tempfile import mkdtemp
from typing import Any, Optional
from uuid import UUID

import numpy as np
from numpy.typing import NDArray
import pyarrow as pa
from loguru import logger

from infrasys.exceptions import ISNotStored
from infrasys.time_series_models import (
    SingleTimeSeries,
    SingleTimeSeriesMetadata,
    TimeSeriesData,
    TimeSeriesMetadata,
)
from infrasys.time_series_storage_base import TimeSeriesStorageBase

EXTENSION = ".arrow"


class ArrowTimeSeriesStorage(TimeSeriesStorageBase):
    """Stores time series in disk"""

    def __init__(self, directory: Path) -> None:
        self._ts_directory = directory

    @classmethod
    def create_with_temp_directory(
        cls, base_directory: Optional[Path] = None
    ) -> "ArrowTimeSeriesStorage":
        """Construct ArrowTimeSeriesStorage with a temporary directory."""
        directory = Path(mkdtemp(dir=base_directory))
        logger.debug("Creating tmp folder at {}", directory)
        atexit.register(clean_tmp_folder, directory)
        return cls(directory)

    @classmethod
    def create_with_permanent_directory(cls, directory: Path) -> "ArrowTimeSeriesStorage":
        """Construct ArrowTimeSeriesStorage with a permanent directory."""
        directory.mkdir(exist_ok=True)
        return cls(directory)

    def get_time_series_directory(self) -> Path:
        return self._ts_directory

    def add_time_series(
        self,
        metadata: TimeSeriesMetadata,
        time_series: TimeSeriesData,
    ) -> None:
        if isinstance(time_series, SingleTimeSeries):
            self.add_raw_single_time_series(metadata.time_series_uuid, time_series.data_array)
        else:
            msg = f"Bug: need to implement add_time_series for {type(time_series)}"
            raise NotImplementedError(msg)

    def add_raw_single_time_series(
        self, time_series_uuid: UUID, time_series_data: NDArray
    ) -> None:
        fpath = self._ts_directory.joinpath(f"{time_series_uuid}{EXTENSION}")
        if not fpath.exists():
            arrow_batch = self._convert_to_record_batch(time_series_data, str(time_series_uuid))
            with pa.OSFile(str(fpath), "wb") as sink:  # type: ignore
                with pa.ipc.new_file(sink, arrow_batch.schema) as writer:
                    writer.write(arrow_batch)
            logger.trace("Saving time series to {}", fpath)
            logger.debug("Added {} to time series storage", time_series_uuid)
        else:
            logger.debug("{} was already stored", time_series_uuid)

    def get_time_series(
        self,
        metadata: TimeSeriesMetadata,
        start_time: datetime | None = None,
        length: int | None = None,
    ) -> Any:
        if isinstance(metadata, SingleTimeSeriesMetadata):
            return self._get_single_time_series(
                metadata=metadata, start_time=start_time, length=length
            )

        msg = f"Bug: need to implement get_time_series for {type(metadata)}"
        raise NotImplementedError(msg)

    def remove_time_series(self, uuid: UUID) -> None:
        fpath = self._ts_directory.joinpath(f"{uuid}{EXTENSION}")
        if not fpath.exists():
            msg = f"No time series with {uuid} is stored"
            raise ISNotStored(msg)
        fpath.unlink()

    def serialize(self, dst: Path | str, src: Optional[Path | str] = None) -> None:
        # From the shutil documentation: the copying operation will continue if
        # it encounters existing directories, and files within the dst tree
        # will be overwritten by corresponding files from the src tree.
        if src is None:
            src = self._ts_directory
        shutil.copytree(src, dst, dirs_exist_ok=True)
        logger.info("Copied time series data to {}", dst)

    def _get_single_time_series(
        self,
        metadata: SingleTimeSeriesMetadata,
        start_time: datetime | None = None,
        length: int | None = None,
    ) -> SingleTimeSeries:
        fpath = self._ts_directory.joinpath(f"{metadata.time_series_uuid}{EXTENSION}")
        with pa.memory_map(str(fpath), "r") as source:
            base_ts = pa.ipc.open_file(source).get_record_batch(0)
            logger.trace("Reading time series from {}", fpath)
        index, length = metadata.get_range(start_time=start_time, length=length)
        columns = base_ts.column_names
        if len(columns) != 1:
            msg = f"Bug: expected a single column: {columns=}"
            raise Exception(msg)
        # This should be equal to metadata.time_series_uuid in versions
        # v0.2.1 or later. Earlier versions used the time series variable name.
        column = columns[0]
        data = base_ts[column][index : index + length]
        if metadata.quantity_metadata is not None:
            np_array = metadata.quantity_metadata.quantity_type(
                data, metadata.quantity_metadata.units
            )
        else:
            np_array = np.array(data)
        return SingleTimeSeries(
            uuid=metadata.time_series_uuid,
            variable_name=metadata.variable_name,
            resolution=metadata.resolution,
            initial_time=start_time or metadata.initial_time,
            data=np_array,
            normalization=metadata.normalization,
        )

    def get_raw_single_time_series(self, time_series_uuid: UUID) -> NDArray:
        fpath = self._ts_directory.joinpath(f"{time_series_uuid}{EXTENSION}")
        with pa.OSFile(str(fpath), "r") as source:
            base_ts = pa.ipc.open_file(source).get_record_batch(0)
            logger.trace("Reading time series from {}", fpath)
        columns = base_ts.column_names
        if len(columns) != 1:
            msg = f"Bug: expected a single column: {columns=}"
            raise Exception(msg)
        column = columns[0]
        return base_ts[column].to_numpy()

    def _convert_to_record_batch(self, time_series_array: NDArray, column: str) -> pa.RecordBatch:
        """Create record batch to save array to disk."""
        pa_array = pa.array(time_series_array)
        schema = pa.schema([pa.field(column, pa_array.type)])
        return pa.record_batch([pa_array], schema=schema)


def clean_tmp_folder(folder: Path | str) -> None:
    shutil.rmtree(folder)
    logger.info("Wiped time series folder: {}", folder)



"""Defines models that can be used for testing the package."""

from typing import Any

from infrasys import Component, Location, System


class SimpleBus(Component):
    """Represents a bus."""

    voltage: float
    coordinates: Location | None = None

    @classmethod
    def example(cls) -> "SimpleBus":
        return SimpleBus(
            name="simple-bus",
            voltage=1.1,
            coordinates=Location(x=0.0, y=0.0),
        )


class GeneratorBase(Component):
    """Base class for generators"""

    available: bool
    bus: SimpleBus


class SimpleGenerator(GeneratorBase):
    """Represents a generator."""

    active_power: float
    rating: float

    @classmethod
    def example(cls) -> "SimpleGenerator":
        return SimpleGenerator(
            name="simple-gen",
            available=True,
            bus=SimpleBus.example(),
            active_power=1.0,
            rating=0.0,
        )


class RenewableGenerator(GeneratorBase):
    """Represents a generator."""

    active_power: float
    rating: float

    @classmethod
    def example(cls) -> "RenewableGenerator":
        return RenewableGenerator(
            name="renewable-gen",
            available=True,
            bus=SimpleBus.example(),
            active_power=1.0,
            rating=0.0,
        )


class SimpleSubsystem(Component):
    """Represents a subsystem."""

    generators: list[SimpleGenerator]

    @classmethod
    def example(cls) -> "SimpleSubsystem":
        return SimpleSubsystem(
            name="simple-subsystem",
            generators=[SimpleGenerator.example()],
        )


class SimpleSystem(System):
    """System used for testing"""

    def __init__(self, my_attr=5, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.data_format_version = "1.0.3"
        self.my_attr = my_attr

    def serialize_system_attributes(self) -> dict[str, Any]:
        return {"my_attr": self.my_attr}

    def deserialize_system_attributes(self, data: dict[str, Any]) -> None:
        self.my_attr = data["my_attr"]


if __name__ == "__main__":
    system = SimpleSystem()
    bus = SimpleBus.example()
    gen = SimpleGenerator.example()
    system.add_components(bus, gen)
    bus2 = system.get_component(SimpleBus, "simple-bus")
    gen2 = SimpleGenerator(name="gen2", active_power=3.0, rating=1.0, bus=bus2, available=True)



"""This module contains base class for handling pint quantity."""

from typing import TYPE_CHECKING, Any, Type

if TYPE_CHECKING:
    from __main__ import BaseQuantity

import numpy as np
import pint
from pydantic import GetCoreSchemaHandler, SerializationInfo
from pydantic_core import core_schema

ureg = pint.UnitRegistry()


class BaseQuantity(ureg.Quantity):  # type: ignore
    """Interface for base quantity."""

    __base_unit__ = None
    _REGISTRY = ureg

    def __new__(cls: Type["BaseQuantity"], value, units=None):
        return super().__new__(cls, value, units)  # type: ignore

    def __init_subclass__(cls, **kwargs):
        if not cls.__base_unit__:
            msg = "__base_unit__ should be defined"
            raise TypeError(msg)
        super().__init_subclass__(**kwargs)

    # Required for pydantic validation
    @classmethod
    def __get_pydantic_core_schema__(
        cls, _: Type["BaseQuantity"], handler: GetCoreSchemaHandler
    ) -> core_schema.CoreSchema:
        return core_schema.with_info_after_validator_function(
            cls._validate,
            core_schema.any_schema(),
            field_name=handler.field_name,
            serialization=core_schema.plain_serializer_function_ser_schema(
                cls._serialize,
                info_arg=True,
                return_schema=core_schema.any_schema(),
            ),
        )

    # Required for pydantic validation
    @classmethod
    def _validate(cls, field_value: Any, _: core_schema.ValidationInfo) -> "BaseQuantity":
        # Type check is more robubst to check that is not an instance of a bare "BaseQuantity"
        if type(field_value) is cls:
            if cls.__base_unit__:
                assert field_value.check(
                    cls.__base_unit__
                ), f"Unit must be compatible with {cls.__base_unit__}"
                return field_value
        if isinstance(field_value, pint.Quantity):
            if cls.__base_unit__:
                assert field_value.check(
                    cls.__base_unit__
                ), f"Unit must be compatible with {cls.__base_unit__}"
                return cls(field_value.magnitude, field_value.units)
        return cls(field_value, cls.__base_unit__)

    @classmethod
    def _serialize(cls, input_value, info: SerializationInfo) -> float | str:
        return_value = input_value
        if context := info.context:  # type: ignore
            # We can add more logic that will change the serialization here.
            magnitude_only = context.get("magnitude_only")
            if magnitude_only:
                return return_value.magnitude
        if info.mode == "json":
            return_value = str(return_value)
        return return_value

    def to_dict(self) -> dict[str, Any]:
        """Convert a quantity to a dictionary for serialization."""
        val = self.magnitude
        if isinstance(self.magnitude, np.ndarray):
            val = self.magnitude.tolist()
        return {"value": val, "units": str(self.units)}

    @classmethod
    def from_dict(cls, data: dict) -> "BaseQuantity":
        """Construct the quantity from a serialized dictionary."""
        return cls(data["value"], data["units"])



"""Manages components"""

import itertools
from collections import defaultdict
from typing import Any, Callable, Iterable, Optional, Type
from uuid import UUID
from loguru import logger

from infrasys.component import Component
from infrasys.component_associations import ComponentAssociations
from infrasys.exceptions import (
    ISAlreadyAttached,
    ISNotStored,
    ISOperationNotAllowed,
)
from infrasys.models import make_label, get_class_and_name_from_label


class ComponentManager:
    """Manages components"""

    def __init__(
        self,
        uuid: UUID,
        auto_add_composed_components: bool,
    ) -> None:
        self._components: dict[Type, dict[str | None, list[Component]]] = {}
        self._components_by_uuid: dict[UUID, Component] = {}
        self._uuid = uuid
        self._auto_add_composed_components = auto_add_composed_components
        self._associations = ComponentAssociations()

    @property
    def auto_add_composed_components(self) -> bool:
        """Return the setting for auto_add_composed_components."""
        return self._auto_add_composed_components

    @auto_add_composed_components.setter
    def auto_add_composed_components(self, val: bool) -> None:
        """Set auto_add_composed_components."""
        self._auto_add_composed_components = val

    def add(self, *components: Component, deserialization_in_progress=False) -> None:
        """Add one or more components to the system.

        Raises
        ------
        ISAlreadyAttached
            Raised if a component is already attached to a system.
        """
        if not components:
            return

        for component in components:
            self._add(component, deserialization_in_progress)

        self._associations.add(*components)

    def get(self, component_type: Type[Component], name: str) -> Any:
        """Return the component with the passed type and name.

        Raises
        ------
        ISDuplicateNames
            Raised if more than one component match the inputs.

        See Also
        --------
        list_by_name
        """
        if component_type not in self._components or name not in self._components[component_type]:
            label = make_label(component_type.__name__, name)
            msg = f"{label} is not stored"
            raise ISNotStored(msg)

        components = self._components[component_type][name]
        if len(components) > 1:
            msg = (
                f"There is more than one {component_type} with {name=}. Please use "
                "list_by_name instead."
            )
            raise ISOperationNotAllowed(msg)

        assert components
        return components[0]

    def get_num_components(self) -> int:
        """Return the number of stored components."""
        return len(self._components_by_uuid)

    def get_num_components_by_type(self) -> dict[Type, int]:
        """Return the number of stored components by type."""
        counts: dict[Type, int] = defaultdict(int)
        for component_type, components_by_type in self._components.items():
            for components_by_name in components_by_type.values():
                counts[component_type] += len(components_by_name)
        return counts

    def get_by_label(self, label: str) -> Any:
        """Return the component with the passed label.

        Raises
        ------
        ISOperationNotAllowed
            Raised if there is more than one matching component.
        """
        class_name, name_or_uuid = get_class_and_name_from_label(label)
        if isinstance(name_or_uuid, UUID):
            return self.get_by_uuid(name_or_uuid)

        for component_type, components_by_name in self._components.items():
            if component_type.__name__ == class_name:
                components = components_by_name.get(name_or_uuid)
                if components is None:
                    msg = f"No component with {label=} is stored."
                    raise ISNotStored(msg)
                if len(components) > 1:
                    msg = f"There is more than one component with {label=}."
                    raise ISOperationNotAllowed(msg)
                return components[0]

        msg = f"No component with {label=} is stored."
        raise ISNotStored(msg)

    def get_types(self) -> Iterable[Type[Component]]:
        """Return an iterable of all stored types."""
        return self._components.keys()

    def has_component(self, component) -> bool:
        """Return True if the component is attached."""
        return component.uuid in self._components_by_uuid

    def iter(
        self, *component_types: Type[Component], filter_func: Callable | None = None
    ) -> Iterable[Any]:
        """Return the components with the passed type and optionally match filter_func.

        If component_type is an abstract type, all matching subtypes will be returned.
        """
        for component_type in component_types:
            yield from self._iter(component_type, filter_func)

    def _iter(
        self, component_type: Type[Component], filter_func: Callable | None
    ) -> Iterable[Any]:
        subclasses = component_type.__subclasses__()
        if subclasses:
            for subclass in subclasses:
                # Recurse.
                yield from self._iter(subclass, filter_func)

        if component_type in self._components:
            if filter_func is None:
                yield from itertools.chain(*self._components[component_type].values())
            else:
                for component in itertools.chain(*self._components[component_type].values()):
                    if filter_func(component):
                        yield component

    def list_by_name(self, component_type: Type[Component], name: str) -> list[Any]:
        """Return all components that match component_type and name.

        The component_type can be an abstract type.
        """
        return list(self.iter(component_type, filter_func=lambda x: x.name == name))

    def get_by_uuid(self, uuid: UUID) -> Any:
        """Return the component with the input UUID.

        Raises
        ------
        ISNotStored
            Raised if the UUID is not stored.
        """
        component = self._components_by_uuid.get(uuid)
        if component is None:
            msg = f"No component with {uuid=} is stored"
            raise ISNotStored(msg)
        return component

    def iter_all(self) -> Iterable[Any]:
        """Return an iterator over all components."""
        return self._components_by_uuid.values()

    def list_child_components(
        self, component: Component, component_type: Optional[Type[Component]] = None
    ) -> list[Component]:
        """Return a list of all components that this component composes."""
        return [
            self.get_by_uuid(x)
            for x in self._associations.list_child_components(
                component, component_type=component_type
            )
        ]

    def list_parent_components(
        self, component: Component, component_type: Optional[Type[Component]] = None
    ) -> list[Component]:
        """Return a list of all components that compose this component."""
        return [
            self.get_by_uuid(x)
            for x in self._associations.list_parent_components(
                component, component_type=component_type
            )
        ]

    def to_records(
        self,
        component_type: Type[Component],
        filter_func: Callable | None = None,
        **kwargs,
    ) -> Iterable[dict]:
        """Return a dictionary representation of the requested components.

        For nested components we only return the label instead of the full component.
        """
        for component in self.iter(component_type, filter_func=filter_func):
            data = component.model_dump(**kwargs)
            for key in data:
                subcomponent = getattr(component, key)
                if issubclass(type(subcomponent), Component):
                    data[key] = subcomponent.label
                elif (
                    isinstance(subcomponent, list)
                    and subcomponent
                    and issubclass(type(subcomponent[0]), Component)
                ):
                    for i, sub_component_ in enumerate(subcomponent):
                        subcomponent[i] = sub_component_.label
            yield data

    def remove(self, component: Component, cascade_down: bool = True, force: bool = False) -> Any:
        """Remove the component from the system and return it.

        Notes
        -----
        Users should not call this directly. It should be called through the system
        so that time series is handled.
        """
        component_type = type(component)
        # The system method should have already performed the check, but for completeness in case
        # someone calls it directly, check here.
        key = component.name or component.label
        if component_type not in self._components or key not in self._components[component_type]:
            msg = f"{component.label} is not stored"
            raise ISNotStored(msg)

        self._check_parent_components_for_remove(component, force)
        container = self._components[component_type][key]
        for i, comp in enumerate(container):
            if comp.uuid == component.uuid:
                container.pop(i)
                if not self._components[component_type][key]:
                    self._components[component_type].pop(key)
                    self._components_by_uuid.pop(component.uuid)
                if not self._components[component_type]:
                    self._components.pop(component_type)
                logger.debug("Removed component {}", component.label)
                if cascade_down:
                    child_components = self._associations.list_child_components(component)
                else:
                    child_components = []
                self._associations.remove(component)
                for child_uuid in child_components:
                    child = self.get_by_uuid(child_uuid)
                    parent_components = self.list_parent_components(child)
                    if not parent_components:
                        self.remove(child, cascade_down=cascade_down, force=force)
                return

        msg = f"Component {component.label} is not stored"
        raise ISNotStored(msg)

    def _check_parent_components_for_remove(self, component: Component, force: bool) -> None:
        parent_components = self.list_parent_components(component)
        if parent_components:
            parent_labels = ", ".join((x.label for x in parent_components))
            if force:
                logger.warning(
                    "Remove {} even though it is attached to these components: {}",
                    component.label,
                    parent_labels,
                )
            else:
                msg = (
                    f"Cannot remove {component.label} because it is attached to these components: "
                    f"{parent_labels}"
                )
                raise ISOperationNotAllowed(msg)

    def copy(
        self,
        component: Component,
        name: str | None = None,
        attach=False,
    ) -> Component:
        """Create a shallow copy of the component."""
        values = {}
        for field in type(component).model_fields:
            cur_val = getattr(component, field)
            if field == "name" and name:
                # Name is special-cased because it is a frozen field.
                val = name
            elif field in ("uuid",):
                continue
            else:
                val = cur_val
            values[field] = val

        new_component = type(component)(**values)  # type: ignore

        logger.info("Copied {} to {}", component.label, new_component.label)
        if attach:
            self.add(new_component)

        return new_component

    def deepcopy(self, component: Component) -> Component:
        """Create a deep copy of the component."""
        values = component.model_dump()
        return type(component)(**values)

    def change_uuid(self, component: Component) -> None:
        """Change the component UUID."""
        msg = "change_component_uuid"
        raise NotImplementedError(msg)

    def rebuild_component_associations(self) -> None:
        """Clear the component associations and rebuild the table. This may be necessary
        if a user reassigns connected components that are part of a system.
        """
        self._associations.clear()
        self._associations.add(*self.iter_all())
        logger.info("Rebuilt all component associations.")

    def update(
        self,
        component_type: Type[Component],
        update_func: Callable,
        filter_func: Callable | None = None,
    ) -> None:
        """Update multiple components of a given type."""

        for component in self.iter(component_type, filter_func=filter_func):
            update_func(component)
        return

    def _add(self, component: Component, deserialization_in_progress: bool) -> None:
        self.raise_if_attached(component)
        if not deserialization_in_progress:
            # TODO: Do we want any checks during deserialization? User could change the JSON.
            # We could prevent the user from changing the JSON with a checksum.
            self._check_component_addition(component)
            component.check_component_addition()
        if component.uuid in self._components_by_uuid:
            msg = f"{component.label} with UUID={component.uuid} is already stored"
            raise ISAlreadyAttached(msg)

        cls = type(component)
        if cls not in self._components:
            self._components[cls] = {}

        name = component.name or component.label
        if name not in self._components[cls]:
            self._components[cls][name] = []

        self._components[cls][name].append(component)
        self._components_by_uuid[component.uuid] = component

        logger.debug("Added {} to the system", component.label)

    def _check_component_addition(self, component: Component) -> None:
        """Check all the fields of a component against the setting
        auto_add_composed_components. Recursive."""
        for field in type(component).model_fields:
            val = getattr(component, field)
            if isinstance(val, Component):
                self._handle_composed_component(val)
                # Recurse.
                self._check_component_addition(val)
            elif isinstance(val, list) and val and isinstance(val[0], Component):
                for item in val:
                    self._handle_composed_component(item)
                    # Recurse.
                    self._check_component_addition(item)

    def _handle_composed_component(self, component: Component) -> None:
        """Do what's needed for a composed component depending on system settings:
        nothing, add, or raise an exception."""
        if component.uuid in self._components_by_uuid:
            return

        if self._auto_add_composed_components:
            logger.debug("Auto-add composed component {}", component.label)
            self._add(component, False)
        else:
            msg = (
                f"Component {component.label} cannot be added to the system because "
                f"its composed component {component.label} is not already attached."
            )
            raise ISOperationNotAllowed(msg)

    def raise_if_attached(self, component: Component):
        """Raise an exception if this component is attached to a system."""
        if component.uuid in self._components_by_uuid:
            msg = f"{component.label} is already attached to the system"
            raise ISAlreadyAttached(msg)

    def raise_if_not_attached(self, component: Component):
        """Raise an exception if this component is not attached to a system.

        Parameters
        ----------
        system_uuid : UUID
            The component must be attached to the system with this UUID.
        """
        if component.uuid not in self._components_by_uuid:
            msg = f"{component.label} is not attached to the system"
            raise ISNotStored(msg)



"""Utility functions for SQLite"""

import sqlite3
from pathlib import Path
from typing import Any, Sequence

from loguru import logger


def backup(src_con: sqlite3.Connection, filename: Path | str) -> None:
    """Backup a database to a file."""
    with sqlite3.connect(filename) as dst_con:
        src_con.backup(dst_con)
    dst_con.close()
    logger.info("Backed up the database to {}.", filename)


def restore(dst_con: sqlite3.Connection, filename: Path | str) -> None:
    """Restore a database from a file."""
    with sqlite3.connect(filename) as src_con:
        src_con.backup(dst_con)
    src_con.close()
    logger.info("Restored the database from {}.", filename)


def create_in_memory_db(database: str = ":memory:") -> sqlite3.Connection:
    """Create an in-memory database."""
    return sqlite3.connect(database)


def execute(cursor: sqlite3.Cursor, query: str, params: Sequence[str] = ()) -> Any:
    """Execute a SQL query."""
    logger.trace("SQL query: {query} {params=}", query)
    return cursor.execute(query, params)



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "Infrastructure Systems"
copyright = "2024, NREL"
author = "NREL"
release = "0.1.0"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "myst_parser",
    "sphinx.ext.githubpages",
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx.ext.todo",
    "sphinx_copybutton",
    "sphinxcontrib.autodoc_pydantic",
    "sphinx_tabs.tabs",
]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

source_suffix = {
    ".txt": "markdown",
    ".md": "markdown",
}

html_theme = "furo"
html_title = "Infrastructure Systems Documentation"
html_theme_options = {
    "navigation_with_keys": True,
}
html_static_path = ["_static"]

todo_include_todos = True
autoclass_content = "both"
autodoc_member_order = "bysource"
todo_include_todos = True
copybutton_only_copy_prompt_lines = True
copybutton_exclude = ".linenos, .gp, .go"
copybutton_line_continuation_character = "\\"
copybutton_here_doc_delimiter = "EOT"
copybutton_prompt_text = "$"
copybutton_copy_empty_lines = False



import itertools
import os
from datetime import timedelta, datetime
from uuid import uuid4

import numpy as np
import pytest

from infrasys.exceptions import (
    ISAlreadyAttached,
    ISNotStored,
    ISOperationNotAllowed,
    ISConflictingArguments,
)
from infrasys import Component, Location, SingleTimeSeries
from infrasys.quantities import ActivePower
from .models.simple_system import (
    GeneratorBase,
    SimpleSystem,
    SimpleBus,
    SimpleGenerator,
    SimpleSubsystem,
    RenewableGenerator,
)


def test_system():
    system = SimpleSystem()
    geo = Location(x=1.0, y=2.0)
    bus = SimpleBus(name="test-bus", voltage=1.1, coordinates=geo)
    gen = SimpleGenerator(name="test-gen", active_power=1.0, rating=1.0, bus=bus, available=True)
    subsystem = SimpleSubsystem(name="test-subsystem", generators=[gen])
    system.add_components(geo, bus, gen, subsystem)
    assert system.add_components() is None  # type: ignore

    gen2 = system.get_component(SimpleGenerator, "test-gen")
    assert gen2 is gen
    assert gen2.bus is bus

    with pytest.raises(ISNotStored):
        system.get_component(SimpleGenerator, "not-present")


def test_system_auto_add_composed_components():
    system = SimpleSystem(auto_add_composed_components=False)
    assert not system.auto_add_composed_components
    geo = Location(x=1.0, y=2.0)
    bus = SimpleBus(name="test-bus", voltage=1.1, coordinates=geo)

    with pytest.raises(ISOperationNotAllowed):
        system.add_component(bus)

    system.auto_add_composed_components = True
    assert system.auto_add_composed_components
    system.add_component(bus)
    assert len(list(system.get_components(Location))) == 1


def test_system_auto_add_composed_components_list():
    system = SimpleSystem(auto_add_composed_components=False)
    assert not system.auto_add_composed_components
    subsystem = SimpleSubsystem.example()
    with pytest.raises(ISOperationNotAllowed):
        system.add_component(subsystem)
    system.auto_add_composed_components = True
    assert system.auto_add_composed_components
    system.add_component(subsystem)


def test_get_components(simple_system: SimpleSystem):
    system = simple_system
    initial_count = 4
    assert len(list(system.get_components(Component))) == initial_count
    system.auto_add_composed_components = True
    for _ in range(5):
        gen = RenewableGenerator.example()
        system.add_component(gen)
    all_components = list(system.get_components(Component))
    # 5 generators, each includes a bus and location
    assert len(all_components) == initial_count + 5 * 3
    generators = list(
        system.get_components(GeneratorBase, filter_func=lambda x: x.name == "renewable-gen")
    )
    assert len(generators) == 5

    with pytest.raises(ISOperationNotAllowed):
        system.get_component(RenewableGenerator, "renewable-gen")

    assert len(list(system.list_components_by_name(RenewableGenerator, "renewable-gen"))) == 5

    gen = all_components[0]
    assert system.get_component_by_uuid(gen.uuid) is gen
    with pytest.raises(ISNotStored):
        system.get_component_by_uuid(uuid4())

    stored_types = sorted((x.__name__ for x in system.get_component_types()))
    assert stored_types == [
        "Location",
        "RenewableGenerator",
        "SimpleBus",
        "SimpleGenerator",
        "SimpleSubsystem",
    ]


def test_get_component_by_label():
    system = SimpleSystem(auto_add_composed_components=True)
    gen = RenewableGenerator.example()
    system.add_component(gen)
    assert system.get_component_by_label(gen.label) is gen
    with pytest.raises(ISNotStored):
        system.get_component_by_label("SimpleGenerator.invalid")
    with pytest.raises(ISNotStored):
        system.get_component_by_label("invalid.invalid")
    coordinates = gen.bus.coordinates
    assert coordinates is not None
    assert system.get_component_by_label(coordinates.label) is coordinates

    gen = RenewableGenerator.example()
    system.add_component(gen)
    with pytest.raises(ISOperationNotAllowed):
        system.get_component_by_label(gen.label)


def test_get_components_multiple_types():
    system = SimpleSystem()
    bus = SimpleBus(name="test-bus", voltage=1.1)
    gen1 = SimpleGenerator(name="gen1", active_power=1.0, rating=1.0, bus=bus, available=True)
    gen2 = SimpleGenerator(name="gen2", active_power=1.0, rating=1.0, bus=bus, available=True)
    gen3 = RenewableGenerator(name="gen2", active_power=1.0, rating=1.0, bus=bus, available=True)
    system.add_components(bus, gen1, gen2, gen3)

    selected_components = list(system.get_components(SimpleBus, SimpleGenerator))
    assert len(selected_components) == 3  # 2 SimpleGenerator + 1 SimpleBus

    # Validate that filter_function works as well
    selected_components = list(
        system.get_components(
            SimpleGenerator, RenewableGenerator, filter_func=lambda x: x.name == "gen2"
        )
    )
    assert len(selected_components) == 2  # 1 SimpleGenerator + 1 RenewableGenerator


def test_component_associations(tmp_path):
    system = SimpleSystem()
    for i in range(3):
        geo = Location(x=i, y=i + 1)
        bus = SimpleBus(name=f"bus{i}", voltage=1.1, coordinates=geo)
        gen1 = SimpleGenerator(
            name=f"gen{i}a", active_power=1.0, rating=1.0, bus=bus, available=True
        )
        gen2 = SimpleGenerator(
            name=f"gen{i}b", active_power=1.0, rating=1.0, bus=bus, available=True
        )
        subsystem = SimpleSubsystem(name=f"test-subsystem{i}", generators=[gen1, gen2])
        system.add_components(geo, bus, gen1, gen2, subsystem)

    def check_attached_components(my_sys):
        for i in range(3):
            bus = my_sys.get_component(SimpleBus, f"bus{i}")
            gen1 = my_sys.get_component(SimpleGenerator, f"gen{i}a")
            gen2 = my_sys.get_component(SimpleGenerator, f"gen{i}b")
            attached = my_sys.list_parent_components(bus, component_type=SimpleGenerator)
            assert len(attached) == 2
            labels = {gen1.label, gen2.label}
            for component in attached:
                assert component.label in labels
                attached_subsystems = my_sys.list_parent_components(component)
                assert len(attached_subsystems) == 1
                assert attached_subsystems[0].name == f"test-subsystem{i}"
                assert not my_sys.list_parent_components(attached_subsystems[0])
                assert my_sys.list_child_components(component) == [bus]
                assert my_sys.list_child_components(component, component_type=SimpleBus) == [bus]

            for component in (bus, gen1, gen2):
                with pytest.raises(ISOperationNotAllowed):
                    my_sys.remove_component(component)

    check_attached_components(system)
    system._component_mgr._associations.clear()
    for component in system.iter_all_components():
        assert not system.list_parent_components(component)

    system.rebuild_component_associations()
    check_attached_components(system)

    save_dir = tmp_path / "test_system"
    system.save(save_dir)
    system2 = SimpleSystem.from_json(save_dir / "system.json")
    check_attached_components(system2)

    bus = system2.get_component(SimpleBus, "bus1")
    with pytest.raises(ISOperationNotAllowed):
        system2.remove_component(bus)
    system2.remove_component(bus, force=True)
    gen = system2.get_component(SimpleGenerator, "gen1a")
    with pytest.raises(ISNotStored):
        system2.get_component(SimpleBus, gen.bus.name)


def test_time_series_attach_from_array():
    system = SimpleSystem()
    bus = SimpleBus(name="test-bus", voltage=1.1)
    gen1 = SimpleGenerator(name="gen1", active_power=1.0, rating=1.0, bus=bus, available=True)
    gen2 = SimpleGenerator(name="gen2", active_power=1.0, rating=1.0, bus=bus, available=True)
    system.add_components(bus, gen1, gen2)
    variable_name = "active_power"
    length = 8784
    data = range(length)
    start = datetime(year=2020, month=1, day=1)
    resolution = timedelta(hours=1)
    ts = SingleTimeSeries.from_array(data, variable_name, start, resolution)
    system.add_time_series(ts, gen1, gen2)
    assert system.has_time_series(gen1, variable_name=variable_name)
    assert system.has_time_series(gen2, variable_name=variable_name)
    assert np.array_equal(system.get_time_series(gen1, variable_name=variable_name).data, ts.data)


def test_time_series():
    system = SimpleSystem()
    bus = SimpleBus(name="test-bus", voltage=1.1)
    gen1 = SimpleGenerator(name="gen1", active_power=1.0, rating=1.0, bus=bus, available=True)
    gen2 = SimpleGenerator(name="gen2", active_power=1.0, rating=1.0, bus=bus, available=True)
    system.add_components(bus, gen1, gen2)

    length = 10
    initial_time = datetime(year=2020, month=1, day=1)
    time_array = [initial_time + timedelta(hours=i) for i in range(length)]
    data = range(length)
    variable_name = "active_power"
    ts = SingleTimeSeries.from_time_array(data, variable_name, time_array)
    with pytest.raises(ValueError, match="The first argument must"):
        # Test a common mistake.
        system.add_time_series(gen1, ts)  # type: ignore

    system.add_time_series(ts, gen1, gen2)
    assert system.has_time_series(gen1, variable_name=variable_name)
    assert system.has_time_series(gen2, variable_name=variable_name)
    assert system.get_time_series(gen1, variable_name=variable_name) == ts
    system.remove_time_series(gen1, gen2, variable_name=variable_name)
    with pytest.raises(ISNotStored):
        system.get_time_series(gen1, variable_name=variable_name)

    assert not system.has_time_series(gen1, variable_name=variable_name)
    assert not system.has_time_series(gen2, variable_name=variable_name)


@pytest.mark.parametrize(
    "in_memory,use_quantity,sql_json",
    list(itertools.product([True, False], [True, False], [True, False])),
)
def test_time_series_retrieval(in_memory, use_quantity, sql_json):
    try:
        if not sql_json:
            os.environ["__INFRASYS_NON_JSON_SQLITE__"] = "1"
        system = SimpleSystem(time_series_in_memory=in_memory)
        bus = SimpleBus(name="test-bus", voltage=1.1)
        gen = SimpleGenerator(name="gen", active_power=1.0, rating=1.0, bus=bus, available=True)
        system.add_components(bus, gen)

        length = 10
        initial_time = datetime(year=2020, month=1, day=1)
        time_array = [initial_time + timedelta(hours=i) for i in range(length)]
        data = (
            [ActivePower(np.random.rand(length), "watts") for _ in range(4)]
            if use_quantity
            else [np.random.rand(length) for _ in range(4)]  # type: ignore
        )
        variable_name = "active_power"
        ts1 = SingleTimeSeries.from_time_array(data[0], variable_name, time_array)
        ts2 = SingleTimeSeries.from_time_array(data[1], variable_name, time_array)
        ts3 = SingleTimeSeries.from_time_array(data[2], variable_name, time_array)
        ts4 = SingleTimeSeries.from_time_array(data[3], variable_name, time_array)
        system.add_time_series(ts1, gen, scenario="high", model_year="2030")
        system.add_time_series(ts2, gen, scenario="high", model_year="2035")
        system.add_time_series(ts3, gen, scenario="low", model_year="2030")
        system.add_time_series(ts4, gen, scenario="low", model_year="2035")
        assert len(system.list_time_series_metadata(gen)) == 4
        assert len(system.list_time_series_metadata(gen, scenario="high", model_year="2035")) == 1
        assert (
            system.list_time_series_metadata(gen, scenario="high", model_year="2035")[
                0
            ].user_attributes["model_year"]
            == "2035"
        )
        assert len(system.list_time_series_metadata(gen, scenario="low")) == 2
        for metadata in system.list_time_series_metadata(gen, scenario="high"):
            assert metadata.user_attributes["scenario"] == "high"

        assert all(
            np.equal(
                system.get_time_series(
                    gen, variable_name, scenario="high", model_year="2030"
                ).data,
                ts1.data,
            )
        )
        assert all(
            np.equal(
                system.get_time_series(
                    gen, variable_name, scenario="high", model_year="2035"
                ).data,
                ts2.data,
            )
        )
        assert all(
            np.equal(
                system.get_time_series(gen, variable_name, scenario="low", model_year="2030").data,
                ts3.data,
            )
        )
        assert all(
            np.equal(
                system.get_time_series(gen, variable_name, scenario="low", model_year="2035").data,
                ts4.data,
            )
        )

        with pytest.raises(ISAlreadyAttached):
            system.add_time_series(ts4, gen, scenario="low", model_year="2035")

        assert system.has_time_series(gen, variable_name=variable_name)
        assert system.has_time_series(gen, variable_name=variable_name, scenario="high")
        assert system.has_time_series(
            gen, variable_name=variable_name, scenario="high", model_year="2030"
        )
        assert not system.has_time_series(gen, variable_name=variable_name, model_year="2036")
        with pytest.raises(ISOperationNotAllowed):
            system.get_time_series(gen, variable_name=variable_name, scenario="high")
        with pytest.raises(ISNotStored):
            system.get_time_series(gen, variable_name=variable_name, scenario="medium")
        assert len(system.list_time_series(gen, variable_name=variable_name, scenario="high")) == 2
        assert len(system.list_time_series(gen, variable_name=variable_name)) == 4
        system.remove_time_series(gen, variable_name=variable_name, scenario="high")
        assert len(system.list_time_series(gen, variable_name=variable_name)) == 2
        system.remove_time_series(gen, variable_name=variable_name)
        assert not system.has_time_series(gen, variable_name=variable_name)
    finally:
        os.environ.pop("__INFRASYS_NON_JSON_SQLITE__", None)


def test_time_series_removal():
    system = SimpleSystem()
    bus = SimpleBus(name="test-bus", voltage=1.1)
    gen1 = SimpleGenerator(name="gen1", active_power=1.0, rating=1.0, bus=bus, available=True)
    gen2 = SimpleGenerator(name="gen2", active_power=1.0, rating=1.0, bus=bus, available=True)
    system.add_components(bus, gen1, gen2)

    variable_names = ["active_power", "reactive_power"]
    uuids = []
    for variable_name in variable_names:
        length = 8784
        data = range(length)
        start = datetime(year=2020, month=1, day=1)
        resolution = timedelta(hours=1)
        ts = SingleTimeSeries.from_array(data, variable_name, start, resolution)
        uuids.append(ts.uuid)
        for gen in (gen1, gen2):
            system.add_time_series(ts, gen, scenario="high", model_year="2030")
            system.add_time_series(ts, gen, scenario="high", model_year="2035")
            system.add_time_series(ts, gen, scenario="low", model_year="2030")
            system.add_time_series(ts, gen, scenario="low", model_year="2035")

    system.remove_time_series(gen1, variable_name="active_power")
    system.remove_time_series(gen1, variable_name="reactive_power")
    assert not system.list_time_series(gen1, variable_name="active_power")
    assert not system.list_time_series(gen1, variable_name="reactive_power")
    assert system.list_time_series(gen2, variable_name="active_power")
    assert system.list_time_series(gen2, variable_name="reactive_power")
    system.remove_time_series(gen2)
    assert not system.list_time_series(gen2, variable_name="active_power")
    assert not system.list_time_series(gen2, variable_name="reactive_power")


def test_time_series_read_only():
    system = SimpleSystem(time_series_read_only=True)
    bus = SimpleBus(name="test-bus", voltage=1.1)
    gen = SimpleGenerator(name="gen", active_power=1.0, rating=1.0, bus=bus, available=True)
    system.add_component(bus)
    system.add_component(gen)

    variable_name = "active_power"
    length = 8784
    data = range(length)
    start = datetime(year=2020, month=1, day=1)
    resolution = timedelta(hours=1)
    ts = SingleTimeSeries.from_array(data, variable_name, start, resolution)
    with pytest.raises(ISOperationNotAllowed):
        system.add_time_series(ts, gen)


def test_serialize_time_series_from_array(tmp_path):
    system = SimpleSystem()
    bus = SimpleBus(name="test-bus", voltage=1.1)
    gen1 = SimpleGenerator(name="gen1", active_power=1.0, rating=1.0, bus=bus, available=True)
    gen2 = SimpleGenerator(name="gen2", active_power=1.0, rating=1.0, bus=bus, available=True)
    system.add_components(bus, gen1, gen2)

    variable_name = "active_power"
    length = 8784
    data = range(length)
    start = datetime(year=2020, month=1, day=1)
    resolution = timedelta(hours=1)
    ts = SingleTimeSeries.from_array(data, variable_name, start, resolution)
    system.add_time_series(ts, gen1, gen2, scenario="high", model_year="2030")
    filename = tmp_path / "system.json"
    system.to_json(filename)

    system2 = SimpleSystem.from_json(filename, time_series_read_only=True)
    gen1b = system2.get_component(SimpleGenerator, gen1.name)
    with pytest.raises(ISOperationNotAllowed):
        system2.remove_time_series(gen1b, variable_name=variable_name)
    ts2 = system.get_time_series(gen1b, variable_name=variable_name)
    assert ts2.data.tolist() == list(data)


@pytest.mark.parametrize("in_memory", [True, False])
def test_time_series_slices(in_memory):
    system = SimpleSystem(
        name="test-system",
        auto_add_composed_components=True,
        time_series_in_memory=in_memory,
    )
    gen = SimpleGenerator.example()
    system.add_components(gen)
    variable_name = "active_power"
    length = 8784
    data = list(range(length))
    start = datetime(year=2020, month=1, day=1)
    resolution = timedelta(hours=1)
    ts = SingleTimeSeries.from_array(data, variable_name, start, resolution)
    system.add_time_series(ts, gen)

    first_timestamp = start
    second_timestamp = start + resolution
    last_timestamp = start + (length - 1) * resolution
    ts_tmp = system.time_series.get(gen, variable_name=variable_name)
    assert isinstance(ts_tmp, SingleTimeSeries)
    assert len(ts_tmp.data) == length
    ts_tmp = system.time_series.get(gen, variable_name=variable_name, length=10)
    assert isinstance(ts_tmp, SingleTimeSeries)
    assert len(ts_tmp.data) == 10
    ts2 = system.time_series.get(
        gen, variable_name=variable_name, start_time=second_timestamp, length=5
    )
    assert isinstance(ts2, SingleTimeSeries)
    assert len(ts2.data) == 5
    assert ts2.data.tolist() == data[1:6]

    ts_tmp = system.time_series.get(gen, variable_name=variable_name, start_time=second_timestamp)
    assert isinstance(ts_tmp, SingleTimeSeries)
    assert len(ts_tmp.data) == len(data) - 1

    with pytest.raises(ISConflictingArguments, match="is less than"):
        system.time_series.get(
            gen,
            variable_name=variable_name,
            start_time=first_timestamp - ts.resolution,
            length=5,
        )
    with pytest.raises(ISConflictingArguments, match="is too large"):
        system.time_series.get(
            gen,
            variable_name=variable_name,
            start_time=last_timestamp + ts.resolution,
            length=5,
        )
    with pytest.raises(ISConflictingArguments, match="conflicts with initial_time"):
        system.time_series.get(
            gen,
            variable_name=variable_name,
            start_time=first_timestamp + timedelta(minutes=1),
        )
    with pytest.raises(ISConflictingArguments, match=r"start_time.*length.*conflicts with"):
        system.time_series.get(
            gen,
            variable_name=variable_name,
            start_time=second_timestamp,
            length=len(data),
        )


def test_copy_component(simple_system_with_time_series: SimpleSystem):
    system = simple_system_with_time_series
    gen1 = system.get_component(SimpleGenerator, "test-gen")

    gen2 = system.copy_component(gen1)
    assert gen2.uuid != gen1.uuid
    assert gen2.name == gen1.name
    assert gen2.bus is gen1.bus

    gen3 = system.copy_component(gen1, name="gen3")
    assert gen3.name == "gen3"

    gen4 = system.copy_component(gen1, name="gen4", attach=True)
    assert gen4.name == "gen4"


def test_deepcopy_component(simple_system_with_time_series: SimpleSystem):
    system = simple_system_with_time_series
    gen1 = system.get_component(SimpleGenerator, "test-gen")
    subsystem = SimpleSubsystem(name="subsystem1", generators=[gen1])
    system.add_component(subsystem)
    gen2 = system.deepcopy_component(gen1)
    assert gen2.name == gen1.name
    assert gen2.uuid == gen1.uuid
    assert gen2.bus.uuid == gen1.bus.uuid
    assert gen2.bus is not gen1.bus


@pytest.mark.parametrize("inputs", [(True, False), (True, False)])
def test_remove_component(inputs):
    in_memory, cascade_down = inputs
    system = SimpleSystem(
        name="test-system",
        auto_add_composed_components=True,
        time_series_in_memory=in_memory,
    )
    gen1 = SimpleGenerator.example()
    bus = gen1.bus
    system.add_components(gen1)
    gen2 = system.copy_component(gen1, name="gen2", attach=True)
    assert gen2.bus is bus
    variable_name = "active_power"
    length = 8784
    data = range(length)
    start = datetime(year=2020, month=1, day=1)
    resolution = timedelta(hours=1)
    ts = SingleTimeSeries.from_array(data, variable_name, start, resolution)
    system.add_time_series(ts, gen1, gen2)

    system.remove_component_by_name(type(gen1), gen1.name, cascade_down=cascade_down)
    assert system.has_component(bus)
    assert not system.has_time_series(gen1)
    assert system.has_time_series(gen2)

    system.remove_component_by_uuid(gen2.uuid, cascade_down=cascade_down)
    assert system.has_component(bus) != cascade_down
    assert not system.has_time_series(gen2)

    with pytest.raises(ISNotStored):
        system.remove_component(gen2)

    with pytest.raises(ISNotStored):
        system.remove_component(gen2)

    for gen in (gen1, gen2):
        with pytest.raises(ISNotStored):
            system.get_component(SimpleGenerator, gen.name)


def test_system_to_dict():
    system = SimpleSystem(
        name="test-system",
        auto_add_composed_components=True,
    )
    gen1 = SimpleGenerator.example()
    gen2 = SimpleGenerator.example()
    gen3 = SimpleGenerator.example()
    system.add_components(gen1, gen2, gen3)

    component_dict: list[dict] = list(system.to_records(SimpleGenerator))
    assert len(component_dict) == 3  # 3 generators
    assert component_dict[0].get("uuid") is not None
    assert component_dict[0]["bus"] == gen1.bus.label

    exclude_first_level_fields = {"name": True, "available": True}
    component_dict = list(system.to_records(SimpleGenerator, exclude=exclude_first_level_fields))
    assert len(component_dict) == 3
    assert component_dict[0].get("name", None) is None
    assert component_dict[0].get("available", None) is None

    component_dict = list(system.to_records(SimpleGenerator))
    assert len(component_dict) == 3  # 3 generators
    assert component_dict[0]["bus"] == gen1.bus.label

    variable_name = "active_power"
    length = 8784
    data = range(length)
    start = datetime(year=2020, month=1, day=1)
    resolution = timedelta(hours=1)
    ts = SingleTimeSeries.from_array(data, variable_name, start, resolution)
    system.add_time_series(ts, gen1)
    component_dicts = list(system.to_records(SimpleGenerator))
    assert len(component_dicts) == 3  # 3 generators


def test_time_series_metadata_sql():
    system = SimpleSystem(name="test-system", auto_add_composed_components=True)
    gen1 = SimpleGenerator.example()
    system.add_components(gen1)
    gen2 = system.copy_component(gen1, name="gen2", attach=True)
    variable_name = "active_power"
    length = 8784
    data = range(length)
    start = datetime(year=2020, month=1, day=1)
    resolution = timedelta(hours=1)
    ts1 = SingleTimeSeries.from_array(data, variable_name, start, resolution)
    ts2 = SingleTimeSeries.from_array(data, variable_name, start, resolution)
    system.add_time_series(ts1, gen1)
    system.add_time_series(ts2, gen2)
    rows = system.time_series.metadata_store.sql(
        f"""
        SELECT component_type, time_series_type, component_uuid, time_series_uuid
        FROM {system.time_series.metadata_store.TABLE_NAME}
        WHERE component_uuid = '{gen1.uuid}'
    """
    )
    assert len(rows) == 1
    row = rows[0]
    assert row[0] == SimpleGenerator.__name__
    assert row[1] == SingleTimeSeries.__name__
    assert row[2] == str(gen1.uuid)
    assert row[3] == str(ts1.uuid)


def test_time_series_metadata_list_rows():
    system = SimpleSystem(name="test-system", auto_add_composed_components=True)
    gen1 = SimpleGenerator.example()
    system.add_components(gen1)
    gen2 = system.copy_component(gen1, name="gen2", attach=True)
    variable_name = "active_power"
    length = 8784
    data = range(length)
    start = datetime(year=2020, month=1, day=1)
    resolution = timedelta(hours=1)
    ts1 = SingleTimeSeries.from_array(data, variable_name, start, resolution)
    ts2 = SingleTimeSeries.from_array(data, variable_name, start, resolution)
    system.add_time_series(ts1, gen1)
    system.add_time_series(ts2, gen2)
    columns = [
        "component_type",
        "time_series_type",
        "component_uuid",
        "time_series_uuid",
    ]
    rows = system.time_series.metadata_store.list_rows(
        gen2,
        variable_name=variable_name,
        time_series_type=SingleTimeSeries.__name__,
        columns=columns,
    )
    assert len(rows) == 1
    row = rows[0]
    assert row[0] == SimpleGenerator.__name__
    assert row[1] == SingleTimeSeries.__name__
    assert row[2] == str(gen2.uuid)
    assert row[3] == str(ts2.uuid)


def test_system_counts():
    system = SimpleSystem(name="test-system", auto_add_composed_components=True)
    gen1 = SimpleGenerator.example()
    gen2 = SimpleGenerator.example()
    system.add_components(gen1, gen2)
    variable_name = "active_power"
    data = range(10)

    def add_time_series(iteration, initial_time, resolution):
        for i in range(5):
            ts1 = SingleTimeSeries.from_array(
                data,
                f"{variable_name}_{iteration}_{i}",
                initial_time + resolution * i,
                resolution,
            )
            ts2 = SingleTimeSeries.from_array(
                data,
                f"{variable_name}_{iteration}_{i}",
                initial_time + resolution * i,
                resolution,
            )
            system.add_time_series(ts1, gen1, gen2)
            system.add_time_series(ts2, gen1.bus)

    add_time_series(1, datetime(year=2020, month=1, day=1), timedelta(hours=1))
    add_time_series(2, datetime(year=2020, month=2, day=1), timedelta(minutes=5))

    # 2 generators, 2 buses, 2 locations
    assert system._components.get_num_components() == 6
    components_by_type = system._components.get_num_components_by_type()
    assert components_by_type[SimpleGenerator] == 2
    assert components_by_type[SimpleBus] == 2
    ts_counts = system.time_series.metadata_store.get_time_series_counts()
    assert ts_counts.time_series_count == 2 * 10
    assert (
        ts_counts.time_series_type_count[
            ("SimpleGenerator", "SingleTimeSeries", "2020-01-01 02:00:00", "1:00:00")
        ]
        == 2
    )
    assert (
        ts_counts.time_series_type_count[
            ("SimpleBus", "SingleTimeSeries", "2020-02-01 00:10:00", "0:05:00")
        ]
        == 1
    )


def test_system_printing(simple_system_with_time_series):
    simple_system_with_time_series.info()



"""Defines the base class for time series storage."""

import abc
from datetime import datetime
from pathlib import Path
from typing import Optional, Any
from uuid import UUID

from infrasys.time_series_models import TimeSeriesData, TimeSeriesMetadata


class TimeSeriesStorageBase(abc.ABC):
    """Base class for time series storage"""

    @abc.abstractmethod
    def add_time_series(self, metadata: TimeSeriesMetadata, time_series: TimeSeriesData) -> None:
        """Store a time series array."""

    @abc.abstractmethod
    def add_raw_single_time_series(self, time_series_uuid: UUID, time_series_data: Any) -> None:
        """Store a time series array from raw data."""

    @abc.abstractmethod
    def get_raw_single_time_series(self, time_series_uuid: UUID) -> Any:
        """Get the raw time series data for a given uuid."""

    @abc.abstractmethod
    def get_time_series_directory(self) -> Path | None:
        """Return the directory containing time series files. Will be none for in-memory time
        series.
        """

    @abc.abstractmethod
    def get_time_series(
        self,
        metadata: TimeSeriesMetadata,
        start_time: datetime | None = None,
        length: int | None = None,
    ) -> TimeSeriesData:
        """Return a time series array."""

    @abc.abstractmethod
    def remove_time_series(self, uuid: UUID) -> None:
        """Remove a time series array and return it."""

    @abc.abstractmethod
    def serialize(self, dst: Path | str, src: Optional[Path | str] = None) -> None:
        """Serialize all time series to the destination directory."""



import logging
from datetime import datetime, timedelta

import pytest
from loguru import logger

from infrasys.location import Location
from infrasys.time_series_models import SingleTimeSeries
from .models.simple_system import SimpleSystem, SimpleBus, SimpleGenerator, SimpleSubsystem


@pytest.fixture
def simple_system() -> SimpleSystem:
    """Creates a system."""
    system = SimpleSystem(name="test-system")
    geo = Location(x=1.0, y=2.0)
    bus = SimpleBus(name="test-bus", voltage=1.1, coordinates=geo)
    gen = SimpleGenerator(name="test-gen", active_power=1.0, rating=1.0, bus=bus, available=True)
    subsystem = SimpleSubsystem(name="test-subsystem", generators=[gen])
    system.add_components(geo, bus, gen, subsystem)
    return system


@pytest.fixture
def simple_system_with_time_series(simple_system) -> SimpleSystem:
    """Creates a system with time series data."""
    variable_name = "active_power"
    length = 8784
    df = range(length)
    start = datetime(year=2020, month=1, day=1)
    resolution = timedelta(hours=1)
    ts = SingleTimeSeries.from_array(df, variable_name, start, resolution)
    gen = simple_system.get_component(SimpleGenerator, "test-gen")
    simple_system.add_time_series(ts, gen)
    return simple_system


@pytest.fixture(autouse=True)
def propagate_logs():
    """Enable logging for the package"""

    class PropagateHandler(logging.Handler):
        def emit(self, record):
            if logging.getLogger(record.name).isEnabledFor(record.levelno):
                logging.getLogger(record.name).handle(record)

    logger.remove()
    logger.enable("infrasys")
    logger.add(PropagateHandler(), format="{message}")
    yield



from pydantic import ValidationError
from infrasys.base_quantity import ureg, BaseQuantity
from infrasys.component import Component
from infrasys.quantities import ActivePower, Time, Voltage
from pint import Quantity
from pint.errors import DimensionalityError
import pytest
import numpy as np


class BaseQuantityComponent(Component):
    voltage: Voltage


def test_base_quantity():
    distance_quantity = ureg.check(None, "meter")(BaseQuantity)

    unit = distance_quantity(100, "meter")
    assert isinstance(unit, BaseQuantity)
    assert isinstance(unit, Quantity)
    # Check that we can not assign units that are not-related.
    with pytest.raises(DimensionalityError):
        _ = distance_quantity(100, "kWh")

    # Check unit multiplication
    active_power_quantity = ActivePower(100, "kW")
    hours = Time(2, "h")

    result_quantity = active_power_quantity * hours
    assert result_quantity.check("[energy]")
    assert result_quantity.magnitude == 200

    # Check to dict
    assert result_quantity.to_dict() == {
        "value": result_quantity.magnitude,
        "units": str(result_quantity.units),
    }


def test_base_quantity_numpy():
    array = np.arange(0, 10)
    measurements = ActivePower(array, "kW")
    assert isinstance(measurements, BaseQuantity)
    assert measurements.to_dict()["value"] == array.tolist()


def test_unit_deserialization():
    test_units = {
        "value": 100,
        "units": "kilowatt",  # The unit name should be the pint default name
    }
    active_power = BaseQuantity.from_dict(test_units)
    assert isinstance(active_power, BaseQuantity)
    assert active_power.magnitude == 100
    assert str(active_power.units) == "kilowatt"


def test_base_unit_validation():
    # Check that new classes must define __base_unit__
    with pytest.raises(TypeError):

        class _(BaseQuantity):
            ...

    test_magnitude = 100
    test_unit = "volt"

    test_quantity = Voltage(test_magnitude, test_unit)

    test_component = BaseQuantityComponent(name="testing", voltage=test_quantity)

    assert test_component.voltage == test_quantity
    assert test_component.voltage.magnitude == test_magnitude
    assert test_component.voltage.units == test_unit

    with pytest.raises(ValidationError):
        BaseQuantityComponent(name="test", voltage=Voltage(test_magnitude, "meter"))

    test_component = BaseQuantityComponent(name="test", voltage=Voltage([0, 1], units="volt"))
    assert type(test_component.voltage) is Voltage
    assert test_component.voltage.magnitude.tolist() == [0, 1]
    assert test_component.voltage.units == test_unit


@pytest.mark.parametrize("input_unit", [Voltage(10, "kV"), 10 * ureg.volt, 10, 10.0])
def test_different_validate(input_unit):
    test_component = BaseQuantityComponent(name="test", voltage=input_unit)
    assert isinstance(test_component.voltage, BaseQuantity)
    assert test_component.voltage.magnitude == 10
    assert test_component.voltage.check(Voltage.__base_unit__)


def test_custom_serialization():
    component = BaseQuantityComponent(name="test", voltage=Voltage(10.0, units="volt"))

    model_dump = component.model_dump(mode="json")

    assert model_dump["voltage"] == str(Voltage(10.0, "volt"))

    model_dump = component.model_dump(context={"magnitude_only": True})
    assert model_dump["voltage"] == 10.0

    model_dump = component.model_dump(mode="json", context={"magnitude_only": True})
    assert model_dump["voltage"] == 10.0



import sqlite3
from typing import Optional, Type
from uuid import UUID

from loguru import logger

from infrasys.component import Component
from infrasys.utils.sqlite import execute


class ComponentAssociations:
    """Stores associations between components. Allows callers to quickly find components composed
    by other components, such as the generator to which a bus is connected."""

    TABLE_NAME = "component_associations"

    def __init__(self) -> None:
        # This uses a different database because it is not persisted when the system
        # is saved to files. It will be rebuilt during de-serialization.
        self._con = sqlite3.connect(":memory:")
        self._create_metadata_table()

    def _create_metadata_table(self):
        schema = [
            "id INTEGER PRIMARY KEY",
            "component_uuid TEXT",
            "component_type TEXT",
            "attached_component_uuid TEXT",
            "attached_component_type TEXT",
        ]
        schema_text = ",".join(schema)
        cur = self._con.cursor()
        execute(cur, f"CREATE TABLE {self.TABLE_NAME}({schema_text})")
        execute(
            cur,
            f"CREATE INDEX by_c_uuid ON {self.TABLE_NAME}(component_uuid, attached_component_uuid)",
        )
        self._con.commit()
        logger.debug("Created in-memory component associations table")

    def add(self, *components: Component):
        """Store an association between each component and directly attached subcomponents.

        - Inspects the type of each field of each component's type. Looks for subtypes of
          Component and lists of subtypes of Component.
        - Does not consider component fields that are dictionaries or other data structures.
        """
        rows = []
        for component in components:
            for field in type(component).model_fields:
                val = getattr(component, field)
                if isinstance(val, Component):
                    rows.append(self._make_row(component, val))
                elif isinstance(val, list) and val and isinstance(val[0], Component):
                    for item in val:
                        rows.append(self._make_row(component, item))

        if rows:
            self._insert_rows(rows)

    def clear(self) -> None:
        """Clear all component associations."""
        execute(self._con.cursor(), f"DELETE FROM {self.TABLE_NAME}")
        logger.info("Cleared all component associations.")

    def list_child_components(
        self, component: Component, component_type: Optional[Type[Component]] = None
    ) -> list[UUID]:
        """Return a list of all component UUIDS that this component composes.
        For example, return the bus attached to a generator.
        """
        where_clause = "WHERE component_uuid = ?"
        if component_type is None:
            params = [str(component.uuid)]
        else:
            params = [str(component.uuid), component_type.__name__]
            where_clause += " AND attached_component_type = ?"
        query = f"SELECT attached_component_uuid FROM {self.TABLE_NAME} {where_clause}"
        cur = self._con.cursor()
        return [UUID(x[0]) for x in execute(cur, query, params)]

    def list_parent_components(
        self, component: Component, component_type: Optional[Type[Component]] = None
    ) -> list[UUID]:
        """Return a list of all component UUIDS that compose this component.
        For example, return all components connected to a bus.
        """
        where_clause = "WHERE attached_component_uuid = ?"
        if component_type is None:
            params = [str(component.uuid)]
        else:
            params = [str(component.uuid), component_type.__name__]
            where_clause += " AND component_type = ?"
        query = f"SELECT component_uuid FROM {self.TABLE_NAME} {where_clause}"
        cur = self._con.cursor()
        return [UUID(x[0]) for x in execute(cur, query, params)]

    def remove(self, component: Component) -> None:
        """Delete all rows with this component."""
        query = f"""
            DELETE
            FROM {self.TABLE_NAME}
            WHERE component_uuid = ? OR attached_component_uuid = ?
        """
        params = [str(component.uuid), str(component.uuid)]
        execute(self._con.cursor(), query, params)
        logger.debug("Removed all associations with component {}", component.label)

    def _insert_rows(self, rows: list[tuple]) -> None:
        cur = self._con.cursor()
        placeholder = ",".join(["?"] * len(rows[0]))
        query = f"INSERT INTO {self.TABLE_NAME} VALUES({placeholder})"
        try:
            cur.executemany(query, rows)
        finally:
            self._con.commit()

    @staticmethod
    def _make_row(component: Component, attached_component: Component):
        return (
            None,
            str(component.uuid),
            type(component).__name__,
            str(attached_component.uuid),
            type(attached_component).__name__,
        )



from typing_extensions import Annotated
from pydantic import Field
from infrasys.models import InfraSysBaseModelWithIdentifers
from infrasys.value_curves import InputOutputCurve, IncrementalCurve, AverageRateCurve, LinearCurve
import pint


class ProductionVariableCostCurve(InfraSysBaseModelWithIdentifers):
    ...


class CostCurve(ProductionVariableCostCurve):
    """Direct representation of the variable operation cost of a power plant in currency.

    Composed of a Value Curve that may represent input-output, incremental, or average rate
    data. The default units for the x-axis are MW and can be specified with
    `power_units`.
    """

    value_curve: Annotated[
        InputOutputCurve | IncrementalCurve | AverageRateCurve,
        Field(
            description="The underlying `ValueCurve` representation of this `ProductionVariableCostCurve`"
        ),
    ]
    vom_cost: Annotated[
        InputOutputCurve,
        Field(description="(default: natural units (MW)) The units for the x-axis of the curve"),
    ] = LinearCurve(0.0)


class FuelCurve(ProductionVariableCostCurve):
    """Representation of the variable operation cost of a power plant in terms of fuel.

    Fuel units (MBTU, liters, m^3, etc.) coupled with a conversion factor between fuel and currency.
    Composed of a Value Curve that may represent input-output, incremental, or average rate data.
    The default units for the x-axis are MW and can be specified with `power_units`.
    """

    value_curve: Annotated[
        InputOutputCurve | IncrementalCurve | AverageRateCurve,
        Field(
            description="The underlying `ValueCurve` representation of this `ProductionVariableCostCurve`"
        ),
    ]
    vom_cost: Annotated[
        InputOutputCurve,
        Field(description="(default: natural units (MW)) The units for the x-axis of the curve"),
    ] = LinearCurve(0.0)
    fuel_cost: Annotated[
        pint.Quantity | float,
        Field(
            description="Either a fixed value for fuel cost or the key to a fuel cost time series"
        ),
    ] = 0.0



"""Defines all exceptions in the package."""


class ISBaseException(Exception):
    """Base class for all exceptions in the package"""


class ISAlreadyAttached(ISBaseException):
    """Raised if the component is already attached to a system."""


class ISDuplicateNames(ISBaseException):
    """Raised if the components with duplicate type and name are stored."""


class ISFileExists(ISBaseException):
    """Raised if the file already exists."""


class ISConflictingArguments(ISBaseException):
    """Raised if the arguments conflict."""


class ISConflictingSystem(ISBaseException):
    """Raised if the system has conflicting values."""


class ISInvalidParameter(ISBaseException):
    """Raised if a parameter is invalid."""


class ISNotStored(ISBaseException):
    """Raised if the requested object is not stored."""


class ISOperationNotAllowed(ISBaseException):
    """Raised if the requested operation is not allowed."""



from infrasys.function_data import (
    LinearFunctionData,
    PiecewiseStepData,
    PiecewiseLinearData,
    XYCoords,
    running_sum,
)
from infrasys import Component
from .models.simple_system import SimpleSystem
import pytest


class FunctionDataComponent(Component):
    function_data: LinearFunctionData


def test_xycoords():
    test_xy = XYCoords(x=1.0, y=2.0)

    # Checking associated types
    assert isinstance(test_xy, XYCoords)

    assert isinstance(test_xy.x, float)

    assert isinstance(test_xy.y, float)


def test_piecewise_linear():
    # Check validation for minimum x values
    test_coords = [XYCoords(1.0, 2.0)]

    with pytest.raises(ValueError):
        PiecewiseLinearData(points=test_coords)

    # Check validation for ascending x values
    test_coords = [XYCoords(1.0, 2.0), XYCoords(4.0, 3.0), XYCoords(3.0, 4.0)]

    with pytest.raises(ValueError):
        PiecewiseLinearData(points=test_coords)


def test_piecewise_step():
    # Check minimum x values
    test_x = [2.0]
    test_y = [1.0]

    with pytest.raises(ValueError):
        PiecewiseStepData(x_coords=test_x, y_coords=test_y)

    # Check ascending x values
    test_x = [1.0, 4.0, 3.0]
    test_y = [2.0, 4.0]

    with pytest.raises(ValueError):
        PiecewiseStepData(x_coords=test_x, y_coords=test_y)

    # Check length of x and y lists
    test_x = [1.0, 2.0, 3.0]
    test_y = [2.0, 4.0, 3.0]

    with pytest.raises(ValueError):
        PiecewiseStepData(x_coords=test_x, y_coords=test_y)


def test_function_data_custom_serialization():
    component = FunctionDataComponent(
        name="test", function_data=LinearFunctionData(proportional_term=1.0, constant_term=2.0)
    )

    model_dump = component.model_dump(mode="json")
    print(model_dump)
    assert model_dump["function_data"]["proportional_term"] == 1.0

    model_dump = component.model_dump(context={"magnitude_only": True})
    assert model_dump["function_data"]["proportional_term"] == 1.0

    model_dump = component.model_dump(mode="json", context={"magnitude_only": True})
    assert model_dump["function_data"]["proportional_term"] == 1.0


def test_function_data_serialization(tmp_path):
    system = SimpleSystem(auto_add_composed_components=True)

    f1 = FunctionDataComponent(
        name="test", function_data=LinearFunctionData(proportional_term=1.0, constant_term=2.0)
    )
    system.add_component(f1)
    filename = tmp_path / "function_data.json"

    system.to_json(filename, overwrite=True)
    system2 = SimpleSystem.from_json(filename)

    assert system2 is not None

    f2 = system2.get_component(FunctionDataComponent, "test")

    assert f2 is not None
    assert f1.function_data.proportional_term == f2.function_data.proportional_term
    assert f1.function_data.constant_term == f2.function_data.constant_term


def test_running_sum():
    test_x = [1.0, 3.0, 6.0]
    test_y = [2.0, 4.0]

    pws = PiecewiseStepData(x_coords=test_x, y_coords=test_y)

    points = running_sum(pws)

    x_values = [p.x for p in points]
    y_values = [p.y for p in points]

    correct_y_values = [0.0, 4.0, 16.0]

    assert x_values == test_x
    assert y_values == correct_y_values



"""Defines models for cost functions"""

from typing import List, NamedTuple

import numpy as np
import pint
from numpy.typing import NDArray
from pydantic import Field, model_validator
from pydantic.functional_validators import AfterValidator
from typing_extensions import Annotated

from infrasys.models import InfraSysBaseModelWithIdentifers


class XYCoords(NamedTuple):
    """Named tuple used to define (x,y) coordinates."""

    x: float
    y: float


class FunctionData(InfraSysBaseModelWithIdentifers):
    """BaseClass of FunctionData"""


class LinearFunctionData(FunctionData):
    """Data representation for linear cost function.

    Used to represent linear cost functions of the form

    .. math:: f(x) = mx + c,

    where :math:`m` is the proportional term and :math:`c` is the constant term.
    """

    proportional_term: Annotated[
        pint.Quantity | float,
        Field(description="the proportional term in the represented function."),
    ]
    constant_term: Annotated[
        float, Field(description="the constant term in the represented function.")
    ]


class QuadraticFunctionData(FunctionData):
    """Data representation for quadratic cost function.

    Used to represent quadratic of cost functions of the form

    .. math:: f(x) = ax^2 + bx + c,

    where :math:`a` is the quadratic term, :math:`b` is the proportional term and :math:`c` is the
    constant term.
    """

    quadratic_term: Annotated[
        float, Field(description="the quadratic term in the represented function.")
    ]
    proportional_term: Annotated[
        float, Field(description="the proportional term in the represented function.")
    ]
    constant_term: Annotated[
        float, Field(description="the constant term in the represented function.")
    ]


def validate_piecewise_linear_x(points: List[XYCoords]) -> List[XYCoords]:
    """Validates the x data for PiecewiseLinearData class

    X data is checked to ensure there is at least two values of x, which is the minimum required to
    generate a cost curve, and is given in ascending order (e.g. [1, 2, 3], not [1, 3, 2]).

    Parameters
    ----------
    points : List[XYCoords]
        List of named tuples of (x,y) coordinates for cost function

    Returns
    -------
    points : List[XYCoords]
        List of (x,y) data for cost function after successful validation.
    """

    x_coords = [p.x for p in points]

    if len(x_coords) < 2:
        msg = "Must specify at least two x-coordinates."
        raise ValueError(msg)
    if not (
        x_coords == sorted(x_coords)
        or (np.isnan(x_coords[0]) and x_coords[1:] == sorted(x_coords[1:]))
    ):
        msg = f"Piecewise x-coordinates must be ascending, got {x_coords}."
        raise ValueError(msg)

    return points


def validate_piecewise_step_x(x_coords: List[float]) -> List[float]:
    """Validates the x data for PiecewiseStepData class

    X data is checked to ensure there is at least two values of x, which is the minimum required to
    generate a cost curve, and is given in ascending order (e.g. [1, 2, 3], not [1, 3, 2]).

    Parameters
    ----------
    x_coords : List[float]
        List of x data for cost function.

    Returns
    -------
    x_coords : List[float]
        List of x data for cost function after successful validation.
    """

    if len(x_coords) < 2:
        msg = "Must specify at least two x-coordinates."
        raise ValueError(msg)
    if not (
        x_coords == sorted(x_coords)
        or (np.isnan(x_coords[0]) and x_coords[1:] == sorted(x_coords[1:]))
    ):
        msg = f"Piecewise x-coordinates must be ascending, got {x_coords}."
        raise ValueError(msg)

    return x_coords


class PiecewiseLinearData(FunctionData):
    """Data representation for piecewise linear cost function.

    Used to represent linear data as a series of points: two points define one segment, three
    points define two segments, etc. The curve starts at the first point given, not the origin.
    Principally used for the representation of cost functions where the points store quantities (x,
    y), such as (MW, USD/h).
    """

    points: Annotated[
        List[XYCoords],
        AfterValidator(validate_piecewise_linear_x),
        Field(description="list of (x,y) points that define the function."),
    ]


class PiecewiseStepData(FunctionData):
    """Data representation for piecewise step cost function.

    Used to represent a step function as a series of endpoint x-coordinates and segment
    y-coordinates: two x-coordinates and one y-coordinate defines a single segment, three
    x-coordinates and two y-coordinates define two segments, etc.

    This can be useful to represent the derivative of a :class:`PiecewiseLinearData`, where the
    y-coordinates of this step function represent the slopes of that piecewise linear function.
    Principally used for the representation of cost functions where the points store quantities (x,
    :math:`dy/dx`), such as (MW, USD/MWh).
    """

    x_coords: Annotated[
        List[float],
        Field(description="the x-coordinates of the endpoints of the segments."),
    ]
    y_coords: Annotated[
        List[float],
        Field(
            description=(
                "The y-coordinates of the segments: `y_coords[1]` is the y-value between "
                "`x_coords[0]` and `x_coords[1]`, etc. Must have one fewer elements than `x_coords`."
            )
        ),
    ]

    @model_validator(mode="after")
    def validate_piecewise_xy(self):
        """Method to validate the x and y data for PiecewiseStepData class

        Model validator used to validate given data for the :class:`PiecewiseStepData`. Calls
        `validate_piecewise_step_x` to check if `x_coords` is valid, then checks if the length of
        `y_coords` is exactly one less than `x_coords`, which is necessary to define the cost
        functions correctly.
        """
        validate_piecewise_step_x(self.x_coords)

        if len(self.y_coords) != len(self.x_coords) - 1:
            msg = "Must specify one fewer y-coordinates than x-coordinates"
            raise ValueError(msg)

        return self


def get_x_lengths(x_coords: List[float]) -> NDArray[np.float64]:
    return np.subtract(x_coords[1:], x_coords[:-1])


def running_sum(data: PiecewiseStepData) -> List[XYCoords]:
    points = []
    slopes = data.y_coords
    x_coords = data.x_coords
    x_lengths = get_x_lengths(x_coords)
    running_y = 0.0

    points.append(XYCoords(x=x_coords[0], y=running_y))
    for prev_slope, this_x, dx in zip(slopes, x_coords[1:], x_lengths):
        running_y += prev_slope * dx
        points.append(XYCoords(x=this_x, y=running_y))

    return points



"""Defines a System"""

import json
import shutil
import sqlite3
from operator import itemgetter
from collections import defaultdict
from datetime import datetime
from pathlib import Path
from typing import Any, Callable, Iterable, Optional, Type
from uuid import UUID, uuid4

from loguru import logger
from rich import print as _pprint
from rich.table import Table

from infrasys.exceptions import (
    ISFileExists,
    ISConflictingArguments,
)
from infrasys.models import make_label
from infrasys.component import (
    Component,
)
from infrasys.component_manager import ComponentManager
from infrasys.serialization import (
    CachedTypeHelper,
    SerializedTypeMetadata,
    SerializedBaseType,
    SerializedComponentReference,
    SerializedQuantityType,
    SerializedType,
    TYPE_METADATA,
)
from infrasys.time_series_manager import TimeSeriesManager, TIME_SERIES_KWARGS
from infrasys.time_series_models import SingleTimeSeries, TimeSeriesData, TimeSeriesMetadata
from infrasys.utils.sqlite import backup, create_in_memory_db, restore


class System:
    """Implements behavior for systems"""

    DB_FILENAME = "time_series_metadata.db"

    def __init__(
        self,
        name: Optional[str] = None,
        description: Optional[str] = None,
        auto_add_composed_components: bool = False,
        con: Optional[sqlite3.Connection] = None,
        time_series_manager: Optional[TimeSeriesManager] = None,
        uuid: Optional[UUID] = None,
        **kwargs: Any,
    ) -> None:
        """Constructs a System.

        Parameters
        ----------
        name : str | None
            Optional system name
        description : str | None
            Optional system description
        auto_add_composed_components : bool
            Set to True to automatically add composed components to the system in add_components.
            The default behavior is to raise an ISOperationNotAllowed when this condition occurs.
            This handles values that are components, such as generator.bus, and lists of
            components, such as subsystem.generators, but not any other form of nested components.
        con : None | sqlite3.Connection
            Users should not pass this. De-serialization (from_json) will pass a Connection.
        time_series_manager : None | TimeSeriesManager
            Users should not pass this. De-serialization (from_json) will pass a constructed
            manager.
        kwargs : Any
            Configures time series behaviors:
              - time_series_in_memory: Defaults to true.
              - time_series_read_only: Disables add/remove of time series, defaults to false.
              - time_series_directory: Location to store time series file, defaults to the system's
                tmp directory.

        Examples
        --------
        >>> system = System(name="my_system")
        >>> system2 = System(name="my_system", time_series_directory="/tmp/scratch")
        """
        self._uuid = uuid or uuid4()
        self._name = name
        self._description = description
        self._component_mgr = ComponentManager(self._uuid, auto_add_composed_components)
        self._con = con or create_in_memory_db()
        time_series_kwargs = {k: v for k, v in kwargs.items() if k in TIME_SERIES_KWARGS}
        self._time_series_mgr = time_series_manager or TimeSeriesManager(
            self._con, **time_series_kwargs
        )
        self._data_format_version: Optional[str] = None
        # Note to devs: if you add new fields, add support in to_json/from_json as appropriate.

        # TODO: add pretty printing of components and time series

    @property
    def auto_add_composed_components(self) -> bool:
        """Return the setting for auto_add_composed_components."""
        return self._component_mgr.auto_add_composed_components

    @auto_add_composed_components.setter
    def auto_add_composed_components(self, val: bool) -> None:
        """Set auto_add_composed_components."""
        self._component_mgr.auto_add_composed_components = val

    def to_json(self, filename: Path | str, overwrite=False, indent=None, data=None) -> None:
        """Write the contents of a system to a JSON file. Time series will be written to a
        directory at the same level as filename.

        Parameters
        ----------
        filename : Path | str
           Filename to write. If the parent directory does not exist, it will be created.
        overwrite : bool
            Set to True to overwrite the file if it already exists.
        indent : int | None
            Indentation level in the JSON file. Defaults to no indentation.
        data : dict | None
            This is an override for packages that compose this System inside a parent System
            class. If set, it will be the outer object in the JSON file. It must not set the
            key 'system'. Packages that derive a custom instance of this class should leave this
            field unset.

        Examples
        --------
        >>> system.to_json("systems/system1.json")
        INFO: Wrote system data to systems/system1.json
        INFO: Copied time series data to systems/system1_time_series
        """
        # TODO: how to get all python package info from environment?
        if isinstance(filename, str):
            filename = Path(filename)
        if filename.exists() and not overwrite:
            msg = f"{filename=} already exists. Choose a different path or set overwrite=True."
            raise ISFileExists(msg)

        filename.parent.mkdir(exist_ok=True)
        time_series_dir = filename.parent / (filename.stem + "_time_series")
        time_series_dir.mkdir(exist_ok=True)
        system_data = {
            "name": self.name,
            "description": self.description,
            "uuid": str(self.uuid),
            "data_format_version": self.data_format_version,
            "components": [x.model_dump_custom() for x in self._component_mgr.iter_all()],
            "time_series": {
                # Note: parent directory is stripped. De-serialization will find it from the
                # parent of the JSON file.
                "directory": time_series_dir.name,
            },
        }
        extra = self.serialize_system_attributes()
        intersection = set(extra).intersection(system_data)
        if intersection:
            msg = f"Extra attributes from parent class collide with System: {intersection}"
            raise ISConflictingArguments(msg)
        system_data.update(extra)

        if data is None:
            data = system_data
        else:
            if "system" in data:
                msg = "data contains the key 'system'"
                raise ISConflictingArguments(msg)
            data["system"] = system_data
        with open(filename, "w", encoding="utf-8") as f_out:
            json.dump(data, f_out, indent=indent)
            logger.info("Wrote system data to {}", filename)

        backup(self._con, time_series_dir / self.DB_FILENAME)
        self._time_series_mgr.serialize(time_series_dir)

    @classmethod
    def from_json(
        cls, filename: Path | str, upgrade_handler: Callable | None = None, **kwargs
    ) -> "System":
        """Deserialize a System from a JSON file. Refer to System constructor for kwargs.

        Parameters
        ----------
        filename : Path | str
            JSON file containing the system data.
        upgrade_handler : Callable | None
            Optional function to handle data format upgrades. Should only be set when the parent
            package composes this package. If set, it will be called before de-serialization of
            the components.

        Examples
        --------
        >>> system = System.from_json("systems/system1.json")
        """
        with open(filename, encoding="utf-8") as f_in:
            data = json.load(f_in)
        time_series_parent_dir = Path(filename).parent
        return cls.from_dict(
            data, time_series_parent_dir, upgrade_handler=upgrade_handler, **kwargs
        )

    def to_records(
        self,
        component_type: Type[Component],
        filter_func: Callable | None = None,
        **kwargs,
    ) -> Iterable[dict]:
        """Return a list of dictionaries of components (records) with the requested type(s) and
        optionally match filter_func.

        Parameters
        ----------
        components:
            Component types to get as dictionaries
        filter_func:
            A function to filter components. Default is None
        kwargs
            Configures Pydantic model_dump behaviour
              - exclude: List or dict of excluded fields.
        Notes
        -----
        If a component type is an abstract type, all matching concrete subtypes will be included in the output.

        It is only recommended to use this function on a single "concrete" types. For example, if
        you have an abstract type called Generator and you create two subtypes called
        ThermalGenerator and RenewableGenerator where some fields are different, if you pass the
        return of System.to_records(Generator) to pandas.DataFrame.from_records, each
        ThermalGenerator row will have NaN values for RenewableGenerator-specific fields.

        Examples
        --------
        To get a tabular representation of a certain type you can use:
        >>> import pandas as pd
        >>> df = pd.DataFrame.from_records(System.to_records(SimpleGen))

        With polars:
        >>> import polars as pl
        >>> df = pl.DataFrame(System.to_records(SimpleGen))

        """
        return self._component_mgr.to_records(component_type, filter_func=filter_func, **kwargs)

    @classmethod
    def from_dict(
        cls,
        data: dict[str, Any],
        time_series_parent_dir: Path | str,
        upgrade_handler: Callable | None = None,
        **kwargs: Any,
    ) -> "System":
        """Deserialize a System from a dictionary.

        Parameters
        ----------
        data : dict[str, Any]
            System data in serialized form.
        time_series_parent_dir : Path | str
            Directory that contains the system's time series directory.
        upgrade_handler : Callable | None
            Optional function to handle data format upgrades. Should only be set when the parent
            package composes this package. If set, it will be called before de-serialization of
            the components.

        Examples
        --------
        >>> system = System.from_dict(data, "systems")
        """
        system_data = data if "system" not in data else data["system"]
        ts_kwargs = {k: v for k, v in kwargs.items() if k in TIME_SERIES_KWARGS}
        ts_path = (
            time_series_parent_dir
            if isinstance(time_series_parent_dir, Path)
            else Path(time_series_parent_dir)
        )
        con = create_in_memory_db()
        restore(con, ts_path / data["time_series"]["directory"] / System.DB_FILENAME)
        time_series_manager = TimeSeriesManager.deserialize(
            con, data["time_series"], ts_path, **ts_kwargs
        )
        system = cls(
            name=system_data.get("name"),
            description=system_data.get("description"),
            con=con,
            time_series_manager=time_series_manager,
            uuid=UUID(system_data["uuid"]),
            **kwargs,
        )
        if system_data.get("data_format_version") != system.data_format_version:
            # This handles the case where the parent package inherited from System.
            system.handle_data_format_upgrade(
                system_data,
                system_data.get("data_format_version"),
                system.data_format_version,
            )
            # This handles the case where the parent package composes an instance of System.
            if upgrade_handler is not None:
                upgrade_handler(
                    system_data,
                    system_data.get("data_format_version"),
                    system.data_format_version,
                )
        system.deserialize_system_attributes(system_data)
        system._deserialize_components(system_data["components"])
        logger.info("Deserialized system {}", system.label)
        return system

    def save(
        self,
        fpath: Path | str,
        filename: str = "system.json",
        zip: bool = False,
        overwrite: bool = False,
    ) -> None:
        """Save the contents of a system and the Time series in a single directory.

        By default, this method creates the user specified folder using the
        `to_json` method. If user sets `zip = True`, we create the folder of
        the user (if it does not exists), zip it to the same location specified
        and delete the folder.

        Parameters
        ----------
        fpath : Path | str
           Filepath to write the contents of the system.
        zip : bool
            Set to True if you want to archive to a zip file.
        filename: str
            Name of the sytem to serialize. Default value: "system.json".
        overwrite: bool
            Overwrites the system if it already exist on the fpath.

        Raises
        ------
        FileExistsError
            Raised if the folder provided exists and the overwrite flag was not provided.

        Examples
        --------
        >>> fpath = Path("folder/subfolder/")
        >>> system.save(fpath)
        INFO: Wrote system data to folder/subfolder/system.json
        INFO: Copied time series data to folder/subfolder/system_time_series

        >>> system_fname = "my_system.json"
        >>> fpath = Path("folder/subfolder/")
        >>> system.save(fpath, filename=system_fname, zip=True)
        INFO: Wrote system data to folder/subfolder/my_system.json
        INFO: Copied time series data to folder/subfolder/my_system_time_series
        INFO: System archived at folder/subfolder/my_system.zip

        See Also
        --------
        to_json: System serialization
        """
        if isinstance(fpath, str):
            fpath = Path(fpath)

        if fpath.exists() and not overwrite:
            msg = f"{fpath} exists already. To overwrite the folder pass `overwrite=True`"
            raise FileExistsError(msg)

        fpath.mkdir(parents=True, exist_ok=True)
        self.to_json(fpath / filename, overwrite=overwrite)

        if zip:
            logger.debug("Archiving system and time series into a single zip file at {}", fpath)
            _ = shutil.make_archive(str(fpath), "zip", fpath)
            logger.debug("Removing {}", fpath)
            shutil.rmtree(fpath)
            logger.info("System archived at {}", fpath)

        return

    def add_component(self, component: Component, **kwargs) -> None:
        """Add one component to the system.

        Parameters
        ----------
        component : Component
            Component to add to the system.

        Raises
        ------
        ISAlreadyAttached
            Raised if a component is already attached to a system.

        Examples
        --------
        >>> system.add_component(Bus.example())

        See Also
        --------
        add_components
        """
        return self.add_components(component, **kwargs)

    def add_components(self, *components: Component, **kwargs) -> None:
        """Add one or more components to the system.

        Parameters
        ----------
        components : Component
            Component(s) to add to the system.

        Raises
        ------
        ISAlreadyAttached
            Raised if a component is already attached to a system.

        Examples
        --------
        >>> system.add_components(Bus.example(), Generator.example())

        See Also
        --------
        add_component
        """
        return self._component_mgr.add(*components, **kwargs)

    def change_component_uuid(self, component: Component) -> None:
        """Change the component UUID. This is required if you copy a component and attach it to
        the same system.

        Parameters
        ----------
        component : Component
        """
        return self._component_mgr.change_uuid(component)

    def copy_component(
        self,
        component: Component,
        name: str | None = None,
        attach: bool = False,
    ) -> Any:
        """Create a copy of the component. Time series data is excluded.

        - The new component will have a different UUID than the original.
        - The copied component will have shared references to any composed components.

        The intention of this method is to provide a way to create variants of a component that
        will be added to the same system. Please refer to :`deepcopy_component`: to create
        copies that are suitable for addition to a different system.

        Parameters
        ----------
        component : Component
            Source component
        name : str
            Optional, if None, keep the original name.
        attach : bool
            Optional, if True, attach the new component to the system.

        Examples
        --------
        >>> gen1 = system.get_component(Generator, "gen1")
        >>> gen2 = system.copy_component(gen, name="gen2")
        >>> gen3 = system.copy_component(gen, name="gen3", attach=True)

        See Also
        --------
        deepcopy_component
        """
        return self._component_mgr.copy(component, name=name, attach=attach)

    def deepcopy_component(self, component: Component) -> Any:
        """Create a deep copy of the component and all composed components. All attributes,
        including names and UUIDs, will be identical to the original. Unlike
        :meth:`copy_component`, there will be no shared references to composed components.

        The intention of this method is to provide a way to create variants of a component that
        will be added to a different system. Please refer to :`copy_component`: to create
        copies that are suitable for addition to the same system.

        Parameters
        ----------
        component : Component
            Source component

        Examples
        --------
        >>> gen1 = system.get_component(Generator, "gen1")
        >>> gen2 = system.deepcopy_component(gen)

        See Also
        --------
        copy_component
        """
        return self._component_mgr.deepcopy(component)

    def get_component(self, component_type: Type[Component], name: str) -> Any:
        """Return the component with the passed type and name.

        Parameters
        ----------
        component_type : Type[Component]
            Type of component
        name : Type
            Name of component

        Raises
        ------
        ISDuplicateNames
            Raised if more than one component match the inputs.

        Examples
        --------
        >>> system.get_component(Generator, "gen1")

        See Also
        --------
        list_by_name
        """
        return self._component_mgr.get(component_type, name)

    def get_component_by_label(self, label: str) -> Any:
        """Return the component with the label.

        Note that this method is slower than :meth:`get_component` because the
        component type cannot be looked up directly. Code that is looping over components
        repeatedly should not use this method.

        Parameters
        ----------
        label : str

        Raises
        ------
        ISNotStored
            Raised if the UUID is not stored.
        ISOperationNotAllowed
            Raised if there is more than one matching component.

        Examples
        --------
        >>> component = system.get_component_by_label("Bus.bus1")
        """
        return self._component_mgr.get_by_label(label)

    def get_component_by_uuid(self, uuid: UUID) -> Any:
        """Return the component with the input UUID.

        Parameters
        ----------
        uuid : UUID

        Raises
        ------
        ISNotStored
            Raised if the UUID is not stored.

        Examples
        --------
        >>> uuid = UUID("714c8311-8dff-4ae2-aa2e-30779a317d42")
        >>> component = system.get_component_by_uuid(uuid)
        """
        return self._component_mgr.get_by_uuid(uuid)

    def get_components(
        self, *component_type: Type[Component], filter_func: Callable | None = None
    ) -> Iterable[Any]:
        """Return the components with the passed type(s) and that optionally match filter_func.

        Parameters
        ----------
        component_type : Type[Component]
            If component_type is an abstract type, all matching subtypes will be returned.
            The function will return all the matching `component_type` passed.
        filter_func : Callable | None
            Optional function to filter the returned values. The function must accept a component
            as a single argument.

        Examples
        --------
        >>> for component in system.get_components(Component)
            print(component.label)
        >>> names = {"bus1", "bus2", "gen1", "gen2"}
        >>> for component in system.get_components(
            Component,
            filter_func=lambda x: x.name in names,
        ):
            print(component.label)

        To request multiple component types:
        >>> for component in system.get_components(SimpleGenerator, SimpleBus)
        print(component.label)
        """
        return self._component_mgr.iter(*component_type, filter_func=filter_func)

    def get_component_types(self) -> Iterable[Type[Component]]:
        """Return an iterable of all component types stored in the system.

        Examples
        --------
        >>> for component_type in system.get_component_types():
        print(component_type)
        """
        return self._component_mgr.get_types()

    def has_component(self, component) -> bool:
        """Return True if the component is attached."""
        return self._component_mgr.has_component(component)

    def list_child_components(
        self, component: Component, component_type: Optional[Type[Component]] = None
    ) -> list[Component]:
        """Return a list of all components that this component composes.

        Parameters
        ----------
        component: Component
        component_type: Optional[Type[Component]]
            Filter the returned list to components of this type.

        See Also
        --------
        list_parent_components
        """
        return self._component_mgr.list_child_components(component, component_type=component_type)

    def list_parent_components(
        self, component: Component, component_type: Optional[Type[Component]] = None
    ) -> list[Component]:
        """Return a list of all components that compose this component.

        An example usage is where you need to find all components connected to a bus and the Bus
        class does not contain that information. The system tracks these connections internally
        and can find those components quickly.

        Parameters
        ----------
        component: Component
        component_type: Optional[Type[Component]]
            Filter the returned list to components of this type.

        Examples
        --------
        >>> components = system.list_parent_components(bus)
        >>> print(f"These components are connected to {bus.label}: ", " ".join(components))

        See Also
        --------
        list_child_components
        """
        return self._component_mgr.list_parent_components(component, component_type=component_type)

    def list_components_by_name(self, component_type: Type[Component], name: str) -> list[Any]:
        """Return all components that match component_type and name.

        Parameters
        ----------
        component_type : Type
        name : str

        Examples
        --------
        system.list_components_by_name(Generator, "gen1")
        """
        return self._component_mgr.list_by_name(component_type, name)

    def iter_all_components(self) -> Iterable[Any]:
        """Return an iterator over all components.

        Examples
        --------
        >>> for component in system.iter_all_components()
            print(component.label)

        See Also
        --------
        get_components
        """
        return self._component_mgr.iter_all()

    def rebuild_component_associations(self) -> None:
        """Clear the component associations and rebuild the table. This may be necessary
        if a user reassigns connected components that are part of a system.
        """
        self._component_mgr.rebuild_component_associations()

    def remove_component(
        self, component: Component, cascade_down: bool = True, force: bool = False
    ) -> Any:
        """Remove the component from the system and return it.

        Parameters
        ----------
        component : Component
        cascade_down : bool
            If True, remove all child components if they have no other parents. Defaults to True.
            For example, if a generator has a bus, no other component holds a reference to that
            bus, and you call remove_component on that generator, the bus will get removed as well.
        force : bool
            If True, remove the component even if other components hold references to this
            component. Defaults to False.

        Raises
        ------
        ISNotStored
            Raised if the component is not stored in the system.
        ISOperationNotAllowed
            Raised if the other components hold references to this component and force=False.

        Examples
        --------
        >>> gen = system.get_component(Generator, "gen1")
        >>> system.remove_component(gen)
        """
        self._component_mgr.raise_if_not_attached(component)
        if self.has_time_series(component):
            for metadata in self._time_series_mgr.list_time_series_metadata(component):
                self.remove_time_series(
                    component,
                    time_series_type=metadata.get_time_series_data_type(),
                    variable_name=metadata.variable_name,
                    **metadata.user_attributes,
                )
        component = self._component_mgr.remove(component, cascade_down=cascade_down, force=force)

    def remove_component_by_name(
        self,
        component_type: Type[Component],
        name: str,
        cascade_down: bool = True,
        force: bool = False,
    ) -> Any:
        """Remove the component with component_type and name from the system and return it.

        Parameters
        ----------
        component_type : Type
        name : str
        cascade_down : bool
            Refer :meth:`remove_component`.
        force : bool
            Refer :meth:`remove_component`.

        Raises
        ------
        ISNotStored
            Raised if the inputs do not match any components in the system.
        ISOperationNotAllowed
            Raised if there is more than one component with component type and name.

        Examples
        --------
        >>> generators = system.remove_by_name(Generator, "gen1")
        """
        component = self.get_component(component_type, name)
        return self.remove_component(component, cascade_down=cascade_down, force=force)

    def remove_component_by_uuid(
        self, uuid: UUID, cascade_down: bool = True, force: bool = False
    ) -> Any:
        """Remove the component with uuid from the system and return it.

        Parameters
        ----------
        uuid : UUID
        cascade_down : bool
            Refer :meth:`remove_component`.
        force : bool
            Refer :meth:`remove_component`.

        Raises
        ------
        ISNotStored
            Raised if the UUID is not stored in the system.

        Examples
        --------
        >>> uuid = UUID("714c8311-8dff-4ae2-aa2e-30779a317d42")
        >>> generator = system.remove_component_by_uuid(uuid)
        """
        component = self.get_component_by_uuid(uuid)
        return self.remove_component(component, cascade_down=cascade_down, force=force)

    def update_components(
        self,
        component_type: Type[Component],
        update_func: Callable,
        filter_func: Callable | None = None,
    ) -> None:
        """Update multiple components of a given type.

        Parameters
        ----------
        component_type : Type[Component]
            Type of component to update. Can be abstract.
        update_func : Callable
            Function to call on each component. Must take a component as a single argument.
        filter_func : Callable | None
            Optional function to filter the components to update. Must take a component as a
            single argument.

        Examples
        --------
        >>> system.update_components(Generator, lambda x: x.active_power *= 10)
        """
        return self._component_mgr.update(component_type, update_func, filter_func=filter_func)

    def add_time_series(
        self,
        time_series: TimeSeriesData,
        *components: Component,
        **user_attributes: Any,
    ) -> None:
        """Store a time series array for one or more components.

        Parameters
        ----------
        time_series : TimeSeriesData
            Time series data to store.
        components : Component
            Add the time series to all of these components.
        user_attributes : Any
            Key/value pairs to store with the time series data. Must be JSON-serializable.

        Raises
        ------
        ISAlreadyAttached
            Raised if the variable name and user attributes match any time series already
            attached to one of the components.
        ISOperationNotAllowed
            Raised if the manager was created in read-only mode.

        Examples
        --------
        >>> gen1 = system.get_component(Generator, "gen1")
        >>> gen2 = system.get_component(Generator, "gen2")
        >>> ts = SingleTimeSeries.from_array(
            data=[0.86, 0.78, 0.81, 0.85, 0.79],
            variable_name="active_power",
            start_time=datetime(year=2030, month=1, day=1),
            resolution=timedelta(hours=1),
        )
        >>> system.add_time_series(ts, gen1, gen2)
        """
        return self._time_series_mgr.add(time_series, *components, **user_attributes)

    def copy_time_series(
        self,
        dst: Component,
        src: Component,
        name_mapping: dict[str, str] | None = None,
    ) -> None:
        """Copy all time series from src to dst.

        Parameters
        ----------
        dst : Component
            Destination component
        src : Component
            Source component
        name_mapping : dict[str, str]
            Optionally map src names to different dst names.
            If provided and src has a time_series with a name not present in name_mapping, that
            time_series will not copied. If name_mapping is nothing then all time_series will be
            copied with src's names.

        Notes
        -----
        name_mapping is currently not implemented.

        Examples
        --------
        >>> gen1 = system.get_component(Generator, "gen1")
        >>> gen2 = system.get_component(Generator, "gen2")
        >>> system.copy_time_series(gen1, gen2)
        """
        return self._time_series_mgr.copy(dst, src, name_mapping=name_mapping)

    def get_time_series(
        self,
        component: Component,
        variable_name: str | None = None,
        time_series_type: Type[TimeSeriesData] = SingleTimeSeries,
        start_time: datetime | None = None,
        length: int | None = None,
        **user_attributes: str,
    ) -> Any:
        """Return a time series array.

        Parameters
        ----------
        component : Component
            Component to which the time series must be attached.
        variable_name : str | None
            Optional, search for time series with this name.
        time_series_type : Type[TimeSeriesData]
            Optional, search for time series with this type.
        start_time : datetime | None
            If not None, take a slice of the time series starting at this time.
        length : int | None
            If not None, take a slice of the time series with this length.
        user_attributes : str
            Optional, search for time series with these attributes.

        Raises
        ------
        ISNotStored
            Raised if no time series matches the inputs.
            Raised if the inputs match more than one time series.
        ISOperationNotAllowed
            Raised if the inputs match more than one time series.

        Examples
        --------
        >>> gen1 = system.get_component(Generator, "gen1")
        >>> ts_full = system.get_time_series(gen1, "active_power")
        >>> ts_slice = system.get_time_series(
            gen1,
            "active_power",
            start_time=datetime(year=2030, month=1, day=1, hour=5),
            length=5,
        )

        See Also
        --------
        list_time_series
        """
        return self._time_series_mgr.get(
            component,
            variable_name=variable_name,
            time_series_type=time_series_type,
            start_time=start_time,
            length=length,
            **user_attributes,
        )

    def has_time_series(
        self,
        component: Component,
        variable_name: Optional[str] = None,
        time_series_type: Type[TimeSeriesData] = SingleTimeSeries,
        **user_attributes: str,
    ) -> bool:
        """Return True if the component has time series matching the inputs.

        Parameters
        ----------
        component : Component
            Component to check for matching time series.
        variable_name : str | None
            Optional, search for time series with this name.
        time_series_type : Type[TimeSeriesData]
            Optional, search for time series with this type.
        user_attributes : str
            Optional, search for time series with these attributes.
        """
        return self.time_series.has_time_series(
            component,
            variable_name=variable_name,
            time_series_type=time_series_type,
            **user_attributes,
        )

    def list_time_series(
        self,
        component: Component,
        variable_name: str | None = None,
        time_series_type: Type[TimeSeriesData] = SingleTimeSeries,
        start_time: datetime | None = None,
        length: int | None = None,
        **user_attributes: Any,
    ) -> list[TimeSeriesData]:
        """Return all time series that match the inputs.

        Parameters
        ----------
        component : Component
            Component to which the time series must be attached.
        variable_name : str | None
            Optional, search for time series with this name.
        time_series_type : Type[TimeSeriesData]
            Optional, search for time series with this type.
        start_time : datetime | None
            If not None, take a slice of the time series starting at this time.
        length : int | None
            If not None, take a slice of the time series with this length.
        user_attributes : str
            Optional, search for time series with these attributes.

        Examples
        --------
        >>> gen1 = system.get_component(Generator, "gen1")
        >>> for ts in system.list_time_series(gen1):
            print(ts)
        """
        return self._time_series_mgr.list_time_series(
            component,
            variable_name=variable_name,
            time_series_type=time_series_type,
            start_time=start_time,
            length=length,
            **user_attributes,
        )

    def list_time_series_metadata(
        self,
        component: Component,
        variable_name: str | None = None,
        time_series_type: Type[TimeSeriesData] = SingleTimeSeries,
        **user_attributes: Any,
    ) -> list[TimeSeriesMetadata]:
        """Return all time series metadata that match the inputs.

        Parameters
        ----------
        component : Component
            Component to which the time series must be attached.
        variable_name : str | None
            Optional, search for time series with this name.
        time_series_type : Type[TimeSeriesData]
            Optional, search for time series with this type.
        user_attributes : str
            Optional, search for time series with these attributes.

        Examples
        --------
        >>> gen1 = system.get_component(Generator, "gen1")
        >>> for metadata in system.list_time_series_metadata(gen1):
            print(metadata)
        """
        return self.time_series.list_time_series_metadata(
            component,
            variable_name=variable_name,
            time_series_type=time_series_type,
            **user_attributes,
        )

    def remove_time_series(
        self,
        *components: Component,
        variable_name: str | None = None,
        time_series_type: Type[TimeSeriesData] = SingleTimeSeries,
        **user_attributes: Any,
    ) -> None:
        """Remove all time series arrays attached to the components matching the inputs.

        Parameters
        ----------
        components : Component
            Affected components
        variable_name : str | None
            Optional, search for time series with this name.
        time_series_type : Type[TimeSeriesData]
            Optional, search for time series with this type.
        user_attributes : str
            Optional, search for time series with these attributes.

        Raises
        ------
        ISNotStored
            Raised if no time series match the inputs.
        ISOperationNotAllowed
            Raised if the manager was created in read-only mode.

        Examples
        --------
        >>> gen1 = system.get_component(Generator, "gen1")
        >>> system.remove_time_series(gen1, "active_power")
        """
        return self._time_series_mgr.remove(
            *components,
            variable_name=variable_name,
            time_series_type=time_series_type,
            **user_attributes,
        )

    def serialize_system_attributes(self) -> dict[str, Any]:
        """Allows subclasses to serialize attributes at the root level."""
        return {}

    def deserialize_system_attributes(self, data: dict[str, Any]) -> None:
        """Allows subclasses to deserialize attributes stored in the JSON at the root level.

        The method should modify self with its custom attributes in data.
        """

    def handle_data_format_upgrade(
        self, data: dict[str, Any], from_version: str | None, to_version: str | None
    ) -> None:
        """Allows subclasses to upgrade data models.

        The parameter data contains the full contents of the serialized JSON file.
        The method should modify the data models in-place.
        """

    def merge_system(self, other: "System") -> None:
        """Merge the contents of another system into this one."""
        msg = "merge_system"
        raise NotImplementedError(msg)

    # TODO: add delete methods that (1) don't raise if not found and (2) don't return anything?

    def convert_storage(self, **kwargs) -> None:
        """
        Converts the time series storage medium.

        Parameters
        ----------
        **kwargs:
            The same keys as TIME_SERIES_KWARGS in time_series_manager.py
            {
                "time_series_in_memory": bool = False,
                "time_series_read_only": bool = False,
                "time_series_directory": Path | None = None,
            }

            Only arguments that need to be changed from the default TIME_SERIES_KWARGS
            need to be passed

        Examples
        --------

        # Initialize the system (defaults to Arrow storage)
        >>> system = infrasys.System(auto_add_composed_components=True)

        # Add components and time series data
        >>> generator, bus, load_data = create_some_data()
        >>> system.add_components(generator, bus)
        >>> system.add_time_series(load_data, generator)

        # Convert the storage to in_memory
        >>> kwargs = {"time_series_in_memory": True}
        >>> system.convert_storage(**kwargs)

        # Check the time series storage type
        >>> isinstance(system._time_series_mgr._storage, InMemoryTimeSeriesStorage)
        True
        """
        return self._time_series_mgr.convert_storage(**kwargs)

    @property
    def _components(self) -> ComponentManager:
        """Return the component manager."""
        return self._component_mgr

    @property
    def data_format_version(self) -> str | None:
        """Return the data format version of the component models."""
        return self._data_format_version

    @data_format_version.setter
    def data_format_version(self, data_format_version: str) -> None:
        """Set the data format version for the component models."""
        self._data_format_version = data_format_version

    @property
    def name(self) -> str | None:
        """Return the name of the system."""
        return self._name

    @name.setter
    def name(self, name: Optional[str]) -> None:
        """Set the name of the system."""
        self._name = name

    @property
    def description(self) -> str | None:
        """Return the description of the system."""
        return self._description

    @description.setter
    def description(self, description: str | None) -> None:
        """Set the description of the system."""
        self._description = description

    @property
    def label(self) -> str:
        """Provides a description of the system."""
        name = self.name or str(self.uuid)
        return make_label(self.__class__.__name__, name)

    @property
    def time_series(self) -> TimeSeriesManager:
        """Return the time series manager."""
        return self._time_series_mgr

    @property
    def uuid(self) -> UUID:
        """Return the UUID of the system."""
        return self._uuid

    def get_time_series_directory(self) -> Path | None:
        """Return the directory containing time series files. Will be none for in-memory time
        series.
        """
        return self.time_series.storage.get_time_series_directory()

    def _deserialize_components(self, components: list[dict[str, Any]]) -> None:
        """Deserialize components from dictionaries and add them to the system."""
        cached_types = CachedTypeHelper()
        skipped_types = self._deserialize_components_first_pass(components, cached_types)
        if skipped_types:
            self._deserialize_components_nested(skipped_types, cached_types)

    def _deserialize_components_first_pass(
        self, components: list[dict], cached_types: CachedTypeHelper
    ) -> dict[Type, list[dict[str, Any]]]:
        deserialized_types = set()
        skipped_types: dict[Type, list[dict[str, Any]]] = defaultdict(list)
        for component_dict in components:
            component = self._try_deserialize_component(component_dict, cached_types)
            if component is None:
                metadata = SerializedTypeMetadata(**component_dict[TYPE_METADATA])
                assert isinstance(metadata.fields, SerializedBaseType)
                component_type = cached_types.get_type(metadata.fields)
                skipped_types[component_type].append(component_dict)
            else:
                deserialized_types.add(type(component))

        cached_types.add_deserialized_types(deserialized_types)
        return skipped_types

    def _deserialize_components_nested(
        self,
        skipped_types: dict[Type, list[dict[str, Any]]],
        cached_types: CachedTypeHelper,
    ) -> None:
        max_iterations = len(skipped_types)
        for _ in range(max_iterations):
            deserialized_types = set()
            for component_type, components in skipped_types.items():
                component = self._try_deserialize_component(components[0], cached_types)
                if component is None:
                    continue
                if len(components) > 1:
                    for component_dict in components[1:]:
                        component = self._try_deserialize_component(component_dict, cached_types)
                        assert component is not None
                deserialized_types.add(component_type)

            for component_type in deserialized_types:
                skipped_types.pop(component_type)
            cached_types.add_deserialized_types(deserialized_types)

        if skipped_types:
            msg = f"Bug: still have types remaining to be deserialized: {skipped_types.keys()}"
            raise Exception(msg)

    def _try_deserialize_component(
        self, component: dict[str, Any], cached_types: CachedTypeHelper
    ) -> Any:
        actual_component = None
        values = self._deserialize_fields(component, cached_types)
        if values is None:
            return None

        metadata = SerializedTypeMetadata(**component[TYPE_METADATA])
        component_type = cached_types.get_type(metadata.fields)
        actual_component = component_type(**values)
        self._components.add(actual_component, deserialization_in_progress=True)
        return actual_component

    def _deserialize_fields(
        self, component: dict[str, Any], cached_types: CachedTypeHelper
    ) -> dict | None:
        values = {}
        for field, value in component.items():
            if isinstance(value, dict) and TYPE_METADATA in value:
                metadata = SerializedTypeMetadata(**value[TYPE_METADATA])
                if isinstance(metadata.fields, SerializedComponentReference):
                    composed_value = self._deserialize_composed_value(
                        metadata.fields, cached_types
                    )
                    if composed_value is None:
                        return None
                    values[field] = composed_value
                elif isinstance(metadata.fields, SerializedQuantityType):
                    quantity_type = cached_types.get_type(metadata.fields)
                    values[field] = quantity_type(value=value["value"], units=value["units"])
                else:
                    msg = f"Bug: unhandled type: {field=} {value=}"
                    raise NotImplementedError(msg)
            elif (
                isinstance(value, list)
                and value
                and isinstance(value[0], dict)
                and TYPE_METADATA in value[0]
                and value[0][TYPE_METADATA]["fields"]["serialized_type"]
                == SerializedType.COMPOSED_COMPONENT.value
            ):
                metadata = SerializedTypeMetadata(**value[0][TYPE_METADATA])
                assert isinstance(metadata.fields, SerializedComponentReference)
                composed_values = self._deserialize_composed_list(value, cached_types)
                if composed_values is None:
                    return None
                values[field] = composed_values
            elif field != TYPE_METADATA:
                values[field] = value

        return values

    def _deserialize_composed_value(
        self, metadata: SerializedComponentReference, cached_types: CachedTypeHelper
    ) -> Any:
        component_type = cached_types.get_type(metadata)
        if cached_types.allowed_to_deserialize(component_type):
            return self._components.get_by_uuid(metadata.uuid)
        return None

    def _deserialize_composed_list(
        self, components: list[dict[str, Any]], cached_types: CachedTypeHelper
    ) -> list[Any] | None:
        deserialized_components = []
        for component in components:
            metadata = SerializedTypeMetadata(**component[TYPE_METADATA])
            assert isinstance(metadata.fields, SerializedComponentReference)
            component_type = cached_types.get_type(metadata.fields)
            if cached_types.allowed_to_deserialize(component_type):
                deserialized_components.append(self._components.get_by_uuid(metadata.fields.uuid))
            else:
                return None
        return deserialized_components

    @staticmethod
    def _make_time_series_directory(filename: Path) -> Path:
        return filename.parent / (filename.stem + "_time_series")

    def show_components(self, component_type):
        # Filtered view of certain concrete types (not really concrete types)
        # We can implement custom printing if we want
        # Dan suggest to remove UUID, system.UUID from component.
        # Nested components gets special handling.
        # What we do with components w/o names? Use .label for nested components.
        raise NotImplementedError

    def info(self):
        info = SystemInfo(system=self)
        info.render()


class SystemInfo:
    """Class to store system component info"""

    def __init__(self, system) -> None:
        self.system = system

    def extract_system_counts(self) -> tuple[int, int, dict, dict]:
        component_count = self.system._components.get_num_components()
        component_type_count = {
            k.__name__: v for k, v in self.system._components.get_num_components_by_type().items()
        }
        ts_counts = self.system.time_series.metadata_store.get_time_series_counts()
        return (
            component_count,
            ts_counts.time_series_count,
            component_type_count,
            ts_counts.time_series_type_count,
        )

    def render(self) -> None:
        """Render Summary information from the system."""
        (
            component_count,
            time_series_count,
            component_type_count,
            time_series_type_count,
        ) = self.extract_system_counts()

        # System table
        system_table = Table(
            title="System",
            show_header=True,
            title_justify="left",
            title_style="bold",
        )
        system_table.add_column("Property")
        system_table.add_column("Value", justify="right")
        system_table.add_row("System name", self.system.name)
        system_table.add_row("Data format version", self.system._data_format_version)
        system_table.add_row("Components attached", f"{component_count}")
        system_table.add_row("Time Series attached", f"{time_series_count}")
        system_table.add_row("Description", self.system.description)
        _pprint(system_table)

        # Component and time series table
        component_table = Table(
            title="Component Information",
            show_header=True,
            title_justify="left",
            title_style="bold",
        )
        component_table.add_column("Type", min_width=20)
        component_table.add_column("Count", justify="right")

        for component_type, component_count in sorted(component_type_count.items()):
            component_table.add_row(
                f"{component_type}",
                f"{component_count}",
            )

        if component_table.rows:
            _pprint(component_table)

        time_series_table = Table(
            title="Time Series Summary",
            show_header=True,
            title_justify="left",
            title_style="bold",
        )
        time_series_table.add_column("Component Type", min_width=20)
        time_series_table.add_column("Time Series Type", justify="right")
        time_series_table.add_column("Initial time", justify="right")
        time_series_table.add_column("Resolution", justify="right")
        time_series_table.add_column("No. Components", justify="right")
        time_series_table.add_column("No. Components with Time Series", justify="right")

        for (
            component_type,
            time_series_type,
            time_series_start_time,
            time_series_resolution,
        ), time_series_count in sorted(time_series_type_count.items(), key=itemgetter(slice(4))):
            time_series_table.add_row(
                f"{component_type}",
                f"{time_series_type}",
                f"{time_series_start_time}",
                f"{time_series_resolution}",
                f"{component_type_count[component_type]}",
                f"{time_series_count}",
            )

        if time_series_table.rows:
            _pprint(time_series_table)



"""Defines models for time series arrays."""

import abc
import importlib
from datetime import datetime, timedelta
from enum import Enum
from typing import (
    Any,
    Literal,
    Optional,
    Type,
    TypeAlias,
    Union,
    Sequence,
)
from uuid import UUID

import numpy as np
import pint
from numpy.typing import NDArray
from pydantic import (
    Field,
    WithJsonSchema,
    field_serializer,
    field_validator,
    computed_field,
    model_validator,
)
from typing_extensions import Annotated

from infrasys.exceptions import (
    ISConflictingArguments,
)
from infrasys.models import InfraSysBaseModelWithIdentifers, InfraSysBaseModel
from infrasys.normalization import NormalizationModel


TIME_COLUMN = "timestamp"
VALUE_COLUMN = "value"


ISArray: TypeAlias = Sequence | NDArray | pint.Quantity


class TimeSeriesStorageType(str, Enum):
    """Defines the possible storage types for time series."""

    HDF5 = "hdf5"
    IN_MEMORY = "in_memory"
    FILE = "arrow"
    PARQUET = "parquet"


class TimeSeriesData(InfraSysBaseModelWithIdentifers, abc.ABC):
    """Base class for all time series models"""

    variable_name: str
    normalization: NormalizationModel = None

    @property
    def summary(self) -> str:
        """Return the variable_name of the time series array with its type."""
        return f"{self.__class__.__name__}.{self.variable_name}"

    @staticmethod
    @abc.abstractmethod
    def get_time_series_metadata_type() -> Type:
        """Return the metadata type associated with this time series type."""


class SingleTimeSeries(TimeSeriesData):
    """Defines a time array with a single dimension of floats."""

    data: NDArray | pint.Quantity
    resolution: timedelta
    initial_time: datetime

    @computed_field
    def length(self) -> int:
        """Return the length of the data."""
        return len(self.data)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, SingleTimeSeries):
            raise NotImplementedError
        is_equal = True
        for field in self.model_fields_set:
            if field == "data":
                if not (self.data == other.data).all():
                    is_equal = False
                    break
            else:
                if not getattr(self, field) == getattr(other, field):
                    is_equal = False
                    break
        return is_equal

    @field_validator("data", mode="before")
    @classmethod
    def check_data(cls, data) -> NDArray | pint.Quantity:  # Standarize what object we receive.
        """Check time series data."""
        if len(data) < 2:
            msg = f"SingleTimeSeries length must be at least 2: {len(data)}"
            raise ValueError(msg)

        if isinstance(data, pint.Quantity):
            if not isinstance(data.magnitude, np.ndarray):
                return type(data)(np.array(data.magnitude), units=data.units)
            return data

        if not isinstance(data, np.ndarray):
            return np.array(data)

        return data

    @classmethod
    def from_array(
        cls,
        data: ISArray,
        variable_name: str,
        initial_time: datetime,
        resolution: timedelta,
        normalization: NormalizationModel = None,
    ) -> "SingleTimeSeries":
        """Method of SingleTimeSeries that creates an instance from a sequence.

        Parameters
        ----------
        data
            Sequence that contains the values of the time series
        initial_time
            Start time for the time series (e.g., datetime(2020,1,1))
        resolution
            Resolution of the time series (e.g., 30min, 1hr)
        variable_name
            Name assigned to the values of the time series (e.g., active_power)

        Returns
        -------
        SingleTimeSeries

        See Also
        --------
        from_time_array:  Time index implementation

        Note
        ----
        - Length of the sequence is inferred from the data.
        """
        if normalization is not None:
            npa = data if isinstance(data, np.ndarray) else np.array(data)
            data = normalization.normalize_array(npa)

        return SingleTimeSeries(
            data=data,  # type: ignore
            variable_name=variable_name,
            initial_time=initial_time,
            resolution=resolution,
            normalization=normalization,
        )

    @classmethod
    def from_time_array(
        cls,
        data: ISArray,
        variable_name: str,
        time_index: Sequence[datetime],
        normalization: NormalizationModel = None,
    ) -> "SingleTimeSeries":
        """Create SingleTimeSeries using time_index provided.

        Parameters
        ----------
        data
            Sequence that contains the values of the time series
        variable_name
            Name assigned to the values of the time series (e.g., active_power)
        time_index
            Sequence that contains the index of the time series

        Returns
        -------
        SingleTimeSeries

        See Also
        --------
        from_array: Base implementation

        Note
        ----
        The current implementation only uses the time_index to infer the initial time and resolution.

        """
        # Infer initial time from the time_index.
        initial_time = time_index[0]

        # This does not cover changes mult-resolution time index.
        resolution = time_index[1] - time_index[0]

        return SingleTimeSeries.from_array(
            data,
            variable_name,
            initial_time,
            resolution,
            normalization=normalization,
        )

    @staticmethod
    def get_time_series_metadata_type() -> Type:
        return SingleTimeSeriesMetadata

    @property
    def data_array(self) -> NDArray:
        if isinstance(self.data, pint.Quantity):
            return self.data.magnitude
        return self.data


class SingleTimeSeriesScalingFactor(SingleTimeSeries):
    """Defines a time array with a single dimension of floats that are 0-1 scaling factors."""


# TODO:
# read CSV and Parquet and convert each column to a SingleTimeSeries


class QuantityMetadata(InfraSysBaseModel):
    """Contains the metadata needed to de-serialize time series stored within a pint.Quantity."""

    module: str
    quantity_type: Annotated[Type, WithJsonSchema({"type": "string"})]
    units: str

    @field_serializer("quantity_type")
    def serialize_type(self, _):
        return self.quantity_type.__name__

    @model_validator(mode="before")
    @classmethod
    def deserialize_from_strings(cls, values: dict[str, Any]) -> dict[str, Any]:
        if isinstance(values["quantity_type"], str):
            module = importlib.import_module(values["module"])
            return {
                "module": values["module"],
                "quantity_type": getattr(module, values["quantity_type"]),
                "units": values["units"],
            }
        return values


class TimeSeriesMetadata(InfraSysBaseModel, abc.ABC):
    """Defines common metadata for all time series."""

    variable_name: str
    initial_time: datetime
    resolution: timedelta
    time_series_uuid: UUID
    user_attributes: dict[str, Any] = {}
    quantity_metadata: Optional[QuantityMetadata] = None
    normalization: NormalizationModel = None
    # TODO: refactor to type_ to avoid overriding builtin?
    type: Literal["SingleTimeSeries", "SingleTimeSeriesScalingFactor"]

    @property
    def label(self) -> str:
        """Return the variable_name of the time series array with its type."""
        return f"{self.type}.{self.variable_name}"

    @staticmethod
    @abc.abstractmethod
    def get_time_series_data_type() -> Type:
        """Return the data type associated with this metadata type."""
        pass

    @staticmethod
    @abc.abstractmethod
    def get_time_series_type_str() -> str:
        """Return the time series type as a string."""


class SingleTimeSeriesMetadataBase(TimeSeriesMetadata, abc.ABC):
    """Base class for SingleTimeSeries metadata."""

    length: int
    type: Literal["SingleTimeSeries", "SingleTimeSeriesScalingFactor"]

    @classmethod
    def from_data(cls, time_series: SingleTimeSeries, **user_attributes) -> Any:
        """Construct a SingleTimeSeriesMetadata from a SingleTimeSeries."""
        quantity_metadata = (
            QuantityMetadata(
                module=type(time_series.data).__module__,
                quantity_type=type(time_series.data),
                units=str(time_series.data.units),
            )
            if isinstance(time_series.data, pint.Quantity)
            else None
        )
        return cls(
            variable_name=time_series.variable_name,
            resolution=time_series.resolution,
            initial_time=time_series.initial_time,
            length=time_series.length,  # type: ignore
            time_series_uuid=time_series.uuid,
            user_attributes=user_attributes,
            quantity_metadata=quantity_metadata,
            normalization=time_series.normalization,
            type=cls.get_time_series_type_str(),  # type: ignore
        )

    def get_range(
        self, start_time: datetime | None = None, length: int | None = None
    ) -> tuple[int, int]:
        """Return the range to be used to index into the dataframe."""
        if start_time is None and length is None:
            return (0, self.length)

        if start_time is None:
            index = 0
        else:
            if start_time < self.initial_time:
                msg = "{start_time=} is less than {self.initial_time=}"
                raise ISConflictingArguments(msg)
            if start_time >= self.initial_time + self.length * self.resolution:
                msg = f"{start_time=} is too large: {self=}"
                raise ISConflictingArguments(msg)
            diff = start_time - self.initial_time
            if (diff % self.resolution).total_seconds() != 0.0:
                msg = (
                    f"{start_time=} conflicts with initial_time={self.initial_time} and "
                    f"resolution={self.resolution}"
                )
                raise ISConflictingArguments(msg)
            index = int(diff / self.resolution)
        if length is None:
            length = self.length - index

        if index + length > self.length:
            msg = f"{start_time=} {length=} conflicts with {self=}"
            raise ISConflictingArguments(msg)

        return (index, length)

    @staticmethod
    def get_time_series_data_type() -> Type:
        return SingleTimeSeries


class SingleTimeSeriesMetadata(SingleTimeSeriesMetadataBase):
    """Defines the metadata for a SingleTimeSeries."""

    type: Literal["SingleTimeSeries"] = "SingleTimeSeries"

    @staticmethod
    def get_time_series_type_str() -> str:
        return "SingleTimeSeries"


class SingleTimeSeriesScalingFactorMetadata(SingleTimeSeriesMetadataBase):
    """Defines the metadata for a SingleTimeSeriesScalingFactor."""

    type: Literal["SingleTimeSeriesScalingFactor"] = "SingleTimeSeriesScalingFactor"

    @staticmethod
    def get_time_series_type_str() -> str:
        return "SingleTimeSeriesScalingFactor"


TimeSeriesMetadataUnion = Annotated[
    Union[SingleTimeSeriesMetadata, SingleTimeSeriesScalingFactorMetadata],
    Field(discriminator="type"),
]



"""Defines models for geographic location."""

from typing_extensions import Annotated

from pydantic import Field

from infrasys import Component


class Location(Component):
    """Specifies geographic location."""

    name: Annotated[str, Field(frozen=True)] = ""
    x: float
    y: float
    crs: str | None = None



import json
import random
import os
from datetime import datetime, timedelta

import numpy as np
import pytest
from pydantic import WithJsonSchema
from typing_extensions import Annotated

from infrasys import Location, SingleTimeSeries
from infrasys.component import Component
from infrasys.quantities import Distance, ActivePower
from infrasys.exceptions import ISOperationNotAllowed
from infrasys.normalization import NormalizationMax
from .models.simple_system import (
    SimpleSystem,
    SimpleBus,
    SimpleGenerator,
    SimpleSubsystem,
)


class ComponentWithPintQuantity(Component):
    """Test component with a container of quantities."""

    distance: Annotated[Distance, WithJsonSchema({"type": "string"})]


def test_serialization(tmp_path):
    system = SimpleSystem(name="test-system", description="a test system", my_attr=5)
    num_components_by_type = 5
    for i in range(num_components_by_type):
        geo = Location(x=random.random(), y=random.random())
        bus = SimpleBus(name=f"test-bus{i}", voltage=random.random(), coordinates=geo)
        gen1 = SimpleGenerator(
            name=f"test-gen{i}a",
            active_power=random.random(),
            rating=random.random(),
            bus=bus,
            available=True,
        )
        gen2 = SimpleGenerator(
            name=f"test-gen{i}b",
            active_power=random.random(),
            rating=random.random(),
            bus=bus,
            available=True,
        )
        subsystem = SimpleSubsystem(name="test-subsystem", generators=[gen1, gen2])
        system.add_components(geo, bus, gen1, gen2, subsystem)

    components = list(system.iter_all_components())
    num_components = len(components)
    assert num_components == num_components_by_type * (1 + 1 + 2 + 1)

    filename = tmp_path / "system.json"
    system.to_json(filename, overwrite=True)
    system2 = SimpleSystem.from_json(filename)
    for key, val in system.__dict__.items():
        if key not in ("_component_mgr", "_time_series_mgr", "_con"):
            assert getattr(system2, key) == val

    components2 = list(system2.iter_all_components())
    assert len(components2) == num_components

    for component in components:
        component2 = system2.get_component_by_uuid(component.uuid)
        for key, val in component.__dict__.items():
            assert getattr(component2, key) == val


@pytest.mark.parametrize("time_series_in_memory", [True, False])
def test_serialize_time_series(tmp_path, time_series_in_memory):
    system = SimpleSystem(time_series_in_memory=time_series_in_memory)
    bus = SimpleBus(name="test-bus", voltage=1.1)
    gen1 = SimpleGenerator(name="gen1", active_power=1.0, rating=1.0, bus=bus, available=True)
    gen2 = SimpleGenerator(name="gen2", active_power=1.0, rating=1.0, bus=bus, available=True)
    system.add_components(bus, gen1, gen2)

    variable_name = "active_power"
    length = 8784
    df = range(length)
    start = datetime(year=2020, month=1, day=1)
    resolution = timedelta(hours=1)
    ts = SingleTimeSeries.from_array(df, variable_name, start, resolution)
    system.add_time_series(ts, gen1, gen2, scenario="high", model_year="2030")
    filename = tmp_path / "system.json"
    system.to_json(filename)

    system2 = SimpleSystem.from_json(filename, time_series_read_only=True)
    system2_ts_dir = system2.get_time_series_directory()
    assert system2_ts_dir is not None
    assert system2_ts_dir == SimpleSystem._make_time_series_directory(filename)
    gen1b = system.get_component(SimpleGenerator, gen1.name)
    with pytest.raises(ISOperationNotAllowed):
        system2.remove_time_series(gen1b, variable_name=variable_name)

    ts2 = system.get_time_series(gen1b, variable_name=variable_name)
    assert np.array_equal(ts2.data, ts.data)

    system3 = SimpleSystem.from_json(filename, time_series_read_only=False)
    assert system3.get_time_series_directory() != SimpleSystem._make_time_series_directory(
        filename
    )
    system3_ts_dir = system3.get_time_series_directory()
    assert system3_ts_dir is not None
    assert not system3_ts_dir.is_relative_to(system2_ts_dir)


@pytest.mark.parametrize(
    "distance",
    [
        Distance(2, "meter"),
        Distance([2, 3], "meter"),
        Distance([[2, 3, 4], [5, 6, 7]], "meter"),
    ],
)
def test_serialize_quantity(tmp_path, distance):
    system = SimpleSystem()
    gen = SimpleGenerator.example()
    component = ComponentWithPintQuantity(name="test", distance=distance)
    assert gen.bus.coordinates is not None
    system.add_components(gen.bus.coordinates, gen.bus, gen, component)
    sys_file = tmp_path / "system.json"
    system.to_json(sys_file)
    system2 = SimpleSystem.from_json(sys_file)
    c1 = system.get_component(ComponentWithPintQuantity, "test")
    c2 = system2.get_component(ComponentWithPintQuantity, "test")
    if isinstance(c1.distance.magnitude, np.ndarray):
        assert (c2.distance == c1.distance).all()
    else:
        assert c2.distance == c1.distance


def test_with_time_series_quantity(tmp_path):
    """Test serialization of SingleTimeSeries with a Pint quantity."""
    system = SimpleSystem(auto_add_composed_components=True)
    gen = SimpleGenerator.example()
    system.add_components(gen)
    length = 10
    initial_time = datetime(year=2020, month=1, day=1)
    resolution = timedelta(hours=1)
    data = ActivePower(range(length), "watts")
    variable_name = "active_power"
    ts = SingleTimeSeries.from_array(data, variable_name, initial_time, resolution)
    system.add_time_series(ts, gen)

    sys_file = tmp_path / "system.json"
    system.to_json(sys_file)

    system2 = SimpleSystem.from_json(sys_file)
    gen2 = system2.get_component(SimpleGenerator, gen.name)
    ts2 = system2.get_time_series(gen2, variable_name=variable_name)
    assert isinstance(ts, SingleTimeSeries)
    assert ts.length == length
    assert ts.resolution == resolution
    assert ts.initial_time == initial_time
    assert isinstance(ts2.data.magnitude, np.ndarray)
    assert np.array_equal(ts2.data.magnitude, np.array(range(length)))


@pytest.mark.parametrize("in_memory", [True, False])
def test_system_with_time_series_normalization(tmp_path, in_memory):
    system = SimpleSystem(
        name="test-system", auto_add_composed_components=True, time_series_in_memory=in_memory
    )
    gen = SimpleGenerator.example()
    system.add_components(gen)
    variable_name = "active_power"
    length = 8784
    data = list(range(length))
    start = datetime(year=2020, month=1, day=1)
    resolution = timedelta(hours=1)
    ts = SingleTimeSeries.from_array(
        data, variable_name, start, resolution, normalization=NormalizationMax()
    )
    system.add_time_series(ts, gen)
    filename = tmp_path / "sys.json"
    system.to_json(filename)

    system2 = SimpleSystem.from_json(filename)
    gen2 = system2.get_component(SimpleGenerator, gen.name)
    ts2 = system2.get_time_series(gen2, variable_name=variable_name)
    assert ts2.normalization.max_value == length - 1


def test_json_schema():
    schema = ComponentWithPintQuantity.model_json_schema()
    assert isinstance(json.loads(json.dumps(schema)), dict)


def test_system_save(tmp_path, simple_system_with_time_series):
    simple_system = simple_system_with_time_series
    custom_folder = "my_system"
    fpath = tmp_path / custom_folder
    fname = "test_system.json"
    simple_system.save(fpath, filename=fname)
    assert os.path.exists(fpath), f"Folder {fpath} was not created successfully"
    assert os.path.exists(fpath / fname), f"Serialized system {fname} was not created successfully"

    with pytest.raises(FileExistsError):
        simple_system.save(fpath, filename=fname)

    simple_system.save(fpath, filename=fname, overwrite=True)
    assert os.path.exists(fpath), f"Folder {fpath} was not created successfully"
    assert os.path.exists(fpath / fname), f"Serialized system {fname} was not created successfully"

    custom_folder = "my_system_zip"
    fpath = tmp_path / custom_folder
    simple_system.save(fpath, filename=fname, zip=True)
    assert not os.path.exists(fpath), f"Original folder {fpath} was not deleted sucessfully."
    zip_fpath = f"{fpath}.zip"
    assert os.path.exists(zip_fpath), f"Zip file {zip_fpath} does not exists"



"""Test related to arrow storage module."""

from datetime import datetime, timedelta

import pytest
import numpy as np

from infrasys.normalization import NormalizationMax
from infrasys.quantities import ActivePower
from infrasys.time_series_models import SingleTimeSeries


def test_single_time_series_attributes():
    start = datetime(year=2020, month=1, day=1)
    resolution = timedelta(hours=1)
    length = 8784
    variable_name = "active_power"
    data = range(length)
    ts = SingleTimeSeries.from_array(
        data=data, variable_name=variable_name, initial_time=start, resolution=resolution
    )
    assert ts.length == length
    assert ts.resolution == resolution
    assert ts.initial_time == start
    assert isinstance(ts.data, np.ndarray)
    assert ts.data[-1] == length - 1


def test_from_array_construction():
    """Test SingleTimeSeries.from_array construction."""
    start = datetime(year=2020, month=1, day=1)
    resolution = timedelta(hours=1)
    length = 8784
    data = range(length)
    variable_name = "active_power"
    ts = SingleTimeSeries.from_array(data, variable_name, start, resolution)
    assert isinstance(ts, SingleTimeSeries)
    assert ts.length == length
    assert ts.resolution == resolution
    assert ts.initial_time == start
    assert isinstance(ts.data, np.ndarray)
    assert ts.data[-1] == length - 1


def test_invalid_sequence_length():
    """Check that time series has at least 2 elements."""
    start = datetime(year=2020, month=1, day=1)
    resolution = timedelta(hours=1)
    length = 1
    data = range(length)
    variable_name = "active_power"
    with pytest.raises(ValueError, match="length must be at least 2"):
        SingleTimeSeries.from_array(data, variable_name, start, resolution)


def test_from_time_array_constructor():
    """Test SingleTimeSeries.from_time_array construction."""
    length = 10
    initial_time = datetime(year=2020, month=1, day=1)
    resolution = timedelta(hours=1)
    time_array = [initial_time + timedelta(hours=i) for i in range(length)]
    data = range(length)
    variable_name = "active_power"
    ts = SingleTimeSeries.from_time_array(data, variable_name, time_array)
    assert isinstance(ts, SingleTimeSeries)
    assert ts.length == length
    assert ts.resolution == resolution
    assert ts.initial_time == initial_time
    assert isinstance(ts.data, np.ndarray)
    assert ts.data[-1] == length - 1


def test_with_quantity():
    """Test SingleTimeSeries with a Pint quantity."""
    length = 10
    initial_time = datetime(year=2020, month=1, day=1)
    resolution = timedelta(hours=1)
    time_array = [initial_time + timedelta(hours=i) for i in range(length)]
    data = ActivePower(range(length), "watts")
    variable_name = "active_power"
    ts = SingleTimeSeries.from_time_array(data, variable_name, time_array)
    assert isinstance(ts, SingleTimeSeries)
    assert ts.length == length
    assert ts.resolution == resolution
    assert ts.initial_time == initial_time
    assert isinstance(ts.data, ActivePower)
    assert ts.data[-1].magnitude == length - 1


def test_normalization():
    length = 10
    initial_time = datetime(year=2020, month=1, day=1)
    time_array = [initial_time + timedelta(hours=i) for i in range(length)]
    data = [1.1, 2.2, 3.3, 4.5, 5.5]
    max_val = data[-1]
    variable_name = "active_power"
    ts = SingleTimeSeries.from_time_array(
        data, variable_name, time_array, normalization=NormalizationMax()
    )
    assert isinstance(ts, SingleTimeSeries)
    assert ts.length == len(data)
    for i, val in enumerate(ts.data):
        assert val == data[i] / max_val



"""Defines base models for components."""

from typing import Any

from pydantic import Field
from rich import print as _pprint
from typing_extensions import Annotated

from infrasys.base_quantity import BaseQuantity
from infrasys.models import (
    InfraSysBaseModelWithIdentifers,
)
from infrasys.serialization import (
    SerializedTypeMetadata,
    SerializedComponentReference,
    SerializedQuantityType,
    TYPE_METADATA,
    serialize_value,
)


class Component(InfraSysBaseModelWithIdentifers):
    """Base class for all models representing entities that get attached to a System."""

    name: Annotated[str, Field(frozen=True)]

    def check_component_addition(self) -> None:
        """Perform checks on the component before adding it to a system."""

    def model_dump_custom(self, *args, **kwargs) -> dict[str, Any]:
        """Custom serialization for this package"""
        refs = {}
        for x in self.model_fields:
            val = self._model_dump_field(x)
            if val is not None:
                refs[x] = val
        exclude = kwargs.get("exclude", [])
        exclude += list(set(exclude).union(refs))
        kwargs["exclude"] = exclude
        data = serialize_value(self, *args, **kwargs)
        data.update(refs)
        return data

    def _model_dump_field(self, field) -> Any:
        val = getattr(self, field)
        if isinstance(val, Component):
            val = {TYPE_METADATA: serialize_component_reference(val)}
        elif isinstance(val, list) and val and isinstance(val[0], Component):
            val = [{TYPE_METADATA: serialize_component_reference(x)} for x in val]
        elif isinstance(val, BaseQuantity):
            data = val.to_dict()
            data[TYPE_METADATA] = SerializedTypeMetadata(
                fields=SerializedQuantityType(
                    module=val.__module__,
                    type=val.__class__.__name__,
                ),
            ).model_dump()
            val = data
        else:
            val = None
        # TODO: other composite types may need handling.
        # Parent packages can always implement a field_serializer themselves.
        return val

    def pprint(self):
        return _pprint(self)


def serialize_component_reference(component: Component) -> dict[str, Any]:
    """Make a JSON serializable reference to a component."""
    return SerializedTypeMetadata(
        fields=SerializedComponentReference(
            module=component.__module__,
            type=component.__class__.__name__,
            uuid=component.uuid,
        ),
    ).model_dump(by_alias=True)



"""Contains logging configuration data."""

import sys

# Logger printing formats
DEFAULT_FORMAT = "<level>{level}</level>: {message}"
DEBUG_FORMAT = (
    "<green>{time:YYYY-MM-DD HH:mm:ss}</green> | "
    "<level>{level: <7}</level> | "
    "<cyan>{name}:{line}</cyan> | "
    "{message}"
)


def setup_logging(
    filename=None,
    level="DEBUG",
) -> None:
    """Configures logging to file and console.

    Parameters
    ----------
    filename : str | None
        log filename
    level : str, optional
        change defualt level of logging.
    verbose :  bool
        returns additional logging information.
    """
    from loguru import logger

    logger.remove()
    logger.enable("infrasys")
    # logger.enable("resource_monitor")
    logger.add(sys.stderr, level=level, format=DEFAULT_FORMAT)
    if filename:
        logger.add(filename, level=level)


if __name__ == "__main__":
    from infrasys import Component, System

    setup_logging(level="INFO")
    system = System()

    component_1 = Component(name="TestComponent")
    component_2 = Component(name="TestComponent2")

    # Adding components
    system.add_component(component_1)
    system.add_component(component_2)
